# transformations.py
# Written by Thomas Hilder and Francesco Bollati

"""
Contains functions for mapping the results between (t,eta,chi) coordinates and physical coordinates, see Bollati et al. 2021 for details.
"""

import numpy                    as np
import matplotlib.pyplot        as plt
from scipy.integrate        import quad, odeint
from scipy.interpolate      import RectBivariateSpline

# NOTE: contents are intended for internal use and should not be directly accessed by users

# TODO: update code to be faster and more efficient, most of the code in this file is still leftover from the old version (Analytical Kinks)

# wake shape
def _phi_wake(r, Rp, hr, q, cw): 
    """Eq. (4) Bollati et al. 2021
    """
    rr = r / Rp
    return -cw * np.sign(r - Rp) * (1 / hr) * (rr**(q - 0.5) / (q - 0.5) - rr**(q + 1) / (q + 1) - 3 / ((2 * q - 1) * (q + 1)))

# eta coordinate transformation
def _Eta(r, phi, Rp, hr, q, cw):
    """Eq. (14) Bollati et al. 2021
    """
    coeff    = 1.5 / hr
    phi_w    = _mod2pi(_phi_wake(r, Rp, hr, q, cw))
    deltaphi = phi - phi_w

    if deltaphi > np.pi:
        deltaphi = deltaphi - 2 * np.pi

    elif deltaphi < -np.pi:
        deltaphi = deltaphi + 2 * np.pi 

    return coeff * deltaphi

def _Eta_vector(r, phi, Rp, hr, q, cw):
    """Eq. (14) Bollati et al. 2021

    Vectorised version of _Eta.
    Simply replaces the modular arithmetic which involved if statements with
    modulus operators and constant offsets.
    """
    coeff    = 1.5 / hr
    phi_w    = _phi_wake(r, Rp, hr, q, cw) % (2*np.pi)
    deltaphi = (phi - phi_w + np.pi) % (2*np.pi) - np.pi

    return coeff * deltaphi

# modulo 2pi
def _mod2pi(phi):
    if phi >= 0:
        return phi % (2 * np.pi) 

    else:
        if np.abs(phi) < np.pi * 2:
            return phi + 2 * np.pi    

        else:
            phi   = -phi
            resto =  phi % (2 * np.pi)    
            
        return -resto + 2 * np.pi


# integrand of the t coordinate transformation
def _t_integrand(x, q, p):
    rho = 5 * q + p
    w   = rho / 2 - 11 / 4
    return np.abs(1 - x**(1.5))**(1.5) * x**w

# integral for t coordinate transformation
def _t_integral(up, q, p):
    return  quad(_t_integrand, 1, up, args=(q,p))[0]

def _t_vector(rr, Rp, hr, q, p):
    """Equation (43) Rafikov 2002    (Eq. 13 Bollati et al. 2021)
    This is a vectorised version of _t.
    _t computes an integral where the integrand is independent of the radius r.
    The radius r only changes the end point of the integral. Instead of using quad
    for each end point (which is not as easily vectorisable), we can utilise an ODE solver to 
    obtain the result at every end point without doing redundant work.
    """

    # First we will flatten the array, keeping track of the original shape.
    # We will also obtain the index of each element of the sorted array.
    # this will allow us to "unsort" the array at the end.
    shape = rr.shape
    integral_bounds = (rr / Rp).flatten()
    integral_bounds_sorted_indices = np.argsort(integral_bounds)

    # We must divide the integral endpoints into values above one and below one,
    # since the lower integral bound is one, and the ode solver only accepts time
    # values in ascending order.
    # This is effectively dividing the integrals between ones which integrate backwards,
    # and ones which integrate forwards.
    integral_bounds_mask  = integral_bounds >= 1.0
    num_larger = np.sum(integral_bounds_mask)

    # We will allocate the arrays to store the values >= 1 and < 1, prepending
    # the initial condition, 1.0. Then we will sort the array since this is required
    # by the ode solver.
    larger_vals = np.zeros( (num_larger + 1,) , dtype = integral_bounds.dtype)
    larger_vals[0] = 1.0
    larger_vals[1:] = integral_bounds[integral_bounds_mask]
    larger_vals.sort()

    smaller_vals = np.zeros( (integral_bounds.size - num_larger + 1,),
        dtype = integral_bounds.dtype)
    smaller_vals[0] = 1.0
    smaller_vals[1:] = integral_bounds[~integral_bounds_mask]
    smaller_vals.sort()
    smaller_vals = np.flip(smaller_vals)

    # Now we can evaluate the integral using the ode solver.
    rho = 5 * q + p
    w   = rho / 2 - 11 / 4

    odefun = lambda _,x: np.abs(1 - x**(1.5))**(1.5) * x**w
    larger_t = odeint(odefun, np.array([0]), larger_vals)

    # In the case of the lower ones, we are making a linear coordinate 
    # transformation x -> 1 - x to make the values ascending and starting at
    # 0 rather than -1 to avoid internal problems in the solver.
    odefun = lambda _,x: np.abs(1 - (1-x)**(1.5))**(1.5) * (1-x)**w
    smaller_t = odeint(odefun, np.array([0]), 1-smaller_vals)

    # combine the results by excluding the prepended initial condition.
    sorted_results = np.concatenate((
        np.flip(smaller_t.flatten()[1:]), 
        larger_t.flatten()[1:]
    ))

    # "unsort", reshape and apply final calculations on the results.
    module_integral = np.zeros_like(sorted_results)
    module_integral[integral_bounds_sorted_indices] = np.abs(sorted_results)

    coeff = 3 * hr**(-5 / 2) / (2**(5 / 4))
    return np.reshape(coeff * module_integral, shape)


# t coordinate transformation
def _t(r, Rp, hr, q, p):
    """Equation (43) Rafikov 2002    (Eq. 13 Bollati et al. 2021)
    """
    module_integral = np.abs(_t_integral(r / Rp, q, p))
    coeff = 3 * hr**(-5 / 2) / (2**(5 / 4))
    return coeff * module_integral

# g(r) quantity calculation
def _g(r, Rp, hr, q, p):
    """Equation (12) Bollati et al. 2021
    """
    coeff = 2**0.25 * hr**0.5
    term1 = (r / Rp)**(0.5 * (1 - p - 3 * q))
    term2 = np.abs((r / Rp)**(-1.5) - 1)**(-0.5)
    return coeff * term1 * term2

# needed to get radial velocities
#def _Lambda_fu(r, Rp, csp, hr, gamma, q, p):
#    """Eq. (28) Bollati et al. 2021
#    """
#    coeff = 2**0.75 * csp * hr**(-0.5) / (gamma + 1)
#    term1 = np.abs((r / Rp)**(-1.5) - 1)**0.5
#    term2 = (r / Rp)**(0.5 * (p + q - 1))
#    return coeff * term1 * term2
#
## needed to get azimuthal velocities
#def _Lambda_fv(r, Rp, csp, hr, gamma, q, p):
#    """Eq. (29) Bollati et al. 2021
#    """
#    coeff = 2**0.75 * csp * hr**0.5 / (gamma + 1)
#    term1 = np.abs((r / Rp)**(-1.5) - 1)**(-0.5)
#    term2 = (r / Rp)**(0.5 * (p - q - 3))
#    return coeff * term1 * term2

# find chi for a particular grid point, either from Burger's eqn solution or self-similar solution
def _get_chi(
    pphi, 
    rr, 
    time_outer,
    time_inner, 
    eta_outer, 
    eta_inner, 
    eta_tilde_outer, 
    eta_tilde_inner,
    C_outer, 
    C_inner,
    solution_outer, 
    solution_inner, 
    t0_outer, 
    t0_inner,
    tf_outer,
    tf_inner, 
    Rp, 
    x_match, 
    l, 
    cw, 
    hr, 
    q, 
    p,
    t1
):
    # COMPUTATION OF Chi

    if (rr > (Rp - x_match*l)) and (rr <(Rp + x_match*l)): # exclude points inside annulus from linear regime
        return 0.

    # change coordinates of the grid point (rr,pphi) to (t1,eta1)
    #t1_orig = _t(rr, Rp, hr, q, p)

    #if np.abs(t1-t1_orig) / t1_orig > 1e-2:
    #    print(t1-t1_orig)


    eta1 = _Eta(rr, pphi, Rp, hr, q, cw)

    # If the point is in the outer disk, use the outer wake solution
    if (rr - Rp) > 0:

        # use numerical solution before the profile develops N-wave
        if t1 < (tf_outer + t0_outer):   
            index_t = np.argmax((t0_outer + time_outer) > t1)
            grid_t = [t0_outer + time_outer[index_t-1], t0_outer+time_outer[index_t]]

            if eta1 > eta_outer[-1] or eta1 < eta_outer[0]:
                Chi = 0
            else:
                index_eta = np.argmax(eta_outer > eta1)
                grid_eta  = np.array([eta_outer[index_eta-1], eta_outer[index_eta]])
                grid_solution = np.array([
                    [solution_outer[index_eta-1,index_t-1], solution_outer[index_eta-1,index_t]],
                    [solution_outer[index_eta,index_t-1],   solution_outer[index_eta,index_t]]
                    ])

                inter = RectBivariateSpline(grid_eta, grid_t, grid_solution, kx=1, ky=1)
                Chi   = inter(eta1, t1)

        # for large t use the analytical N-wave shape (Eq. 17 Bollati et al. 2021)
        else: 

            extr_left  = +cw * np.sign(rr - Rp) * eta_tilde_outer - np.sqrt(2 * C_outer * (t1 - t0_outer))
            extr_right = +cw * np.sign(rr - Rp) * eta_tilde_outer + np.sqrt(2 * C_outer * (t1 - t0_outer))

            if eta1 > extr_left and eta1 < extr_right:
                Chi = (-cw * np.sign(rr - Rp) * eta1 + eta_tilde_outer) / (t1 - t0_outer)  #eq.(29) nonlinear.pdf
            else:
                Chi = 0

    # If the point is in the inner disk, use the inner wake solution
    else:

        # use numerical solution before the profile develops N-wave
        if t1 < (tf_inner + t0_inner):   
            index_t = np.argmax((t0_inner + time_inner) > t1)
            grid_t = [t0_inner + time_inner[index_t-1], t0_inner+time_inner[index_t]]

            if eta1 > eta_inner[-1] or eta1 < eta_inner[0]:
                Chi = 0
            else:
                index_eta = np.argmax(eta_inner > eta1)
                grid_eta  = np.array([eta_inner[index_eta-1], eta_inner[index_eta]])
                grid_solution = np.array([
                    [solution_inner[index_eta-1,index_t-1], solution_inner[index_eta-1,index_t]],
                    [solution_inner[index_eta,index_t-1],   solution_inner[index_eta,index_t]]
                    ])

                inter = RectBivariateSpline(grid_eta, grid_t, grid_solution, kx=1, ky=1)
                Chi   = -inter(eta1, t1)

        # for large t use the analytical N-wave shape (Eq. 17 Bollati et al. 2021)
        else: 

            extr_left  = +cw * np.sign(rr - Rp) * eta_tilde_inner - np.sqrt(2 * C_inner * (t1 - t0_inner))
            extr_right = +cw * np.sign(rr - Rp) * eta_tilde_inner + np.sqrt(2 * C_inner * (t1 - t0_inner))

            if eta1 > extr_left and eta1 < extr_right:
                Chi = (-cw * np.sign(rr - Rp) * eta1 + eta_tilde_inner) / (t1 - t0_inner)  #eq.(29) nonlinear.pdf
            else:
                Chi = 0
    
    return Chi


def _get_chi_vector(
    pphi, 
    rr,
    tt,
    time_outer,
    time_inner, 
    eta_outer, 
    eta_inner, 
    eta_tilde_outer, 
    eta_tilde_inner,
    C_outer, 
    C_inner,
    solution_outer, 
    solution_inner, 
    t0_outer, 
    t0_inner,
    tf_outer,
    tf_inner, 
    Rp, 
    x_match, 
    l, 
    cw, 
    hr, 
    q, 
    p,
): 
    """
    This is a vectorised version of the previous _get_chi function.
    Key changes are:
     - Using masks rather than logical statements to replace
        if statements.
     - Using one RectBivariateSpline interpolation function for all
        points needing interpolation, rather than recreating a function
        for every point.
    """
 # COMPUTATION OF Chi

    Chi = np.zeros_like(rr)

    eta_array = _Eta_vector(rr, pphi, Rp, hr, q, cw)

    # Inner and outer masks will account for the annulus directly.
    outer_mask = rr - Rp >= x_match * l
    inner_mask = rr - Rp <= -x_match * l

    before_N_wave_mask = np.logical_or(
        tt < (tf_outer + t0_outer),
        tt < (tf_inner + t0_inner)
        )

    # Outer masks.

    """
    if (rr - Rp) >= x_match*l:
        if t1 >= (tf_outer + t0_outer):
            if eta1 > extr_left and eta1 < extr_right:
    """
    m = np.logical_and(outer_mask,
        np.logical_and(~before_N_wave_mask,
        np.abs( eta_array - cw * np.sign(rr - Rp) * eta_tilde_outer) > np.sqrt(2 * C_outer * np.abs(tt - t0_outer))
        ))
    Chi[m] = 0.0

    """
    if (rr - Rp) >= x_match*l:
        if t1 >= (tf_outer + t0_outer):
            if eta1 <= extr_left or eta1 >= extr_right:
    """
    m = np.logical_and(outer_mask,
        np.logical_and(~before_N_wave_mask,
        np.abs( eta_array - cw * np.sign(rr - Rp) * eta_tilde_outer) <= np.sqrt(2 * C_outer * np.abs(tt - t0_outer))
        ))
    Chi[m] = (-cw * np.sign(rr[m] - Rp) * eta_array[m] + eta_tilde_outer) / (tt[m] - t0_outer)


    """
    if (rr - Rp) >= x_match*l:
        if t1 < (tf_outer + t0_outer):
            if eta1 > eta_outer[-1] or eta1 < eta_outer[0]:
    """
    interp_outer = RectBivariateSpline(eta_outer, t0_outer + time_outer, solution_outer, kx=1, ky=1)
    m = np.logical_and(outer_mask, 
        np.logical_and(before_N_wave_mask,
        np.logical_and(eta_outer[0] < eta_array,
            eta_array < eta_outer[-1]
        )))
    Chi[m] = interp_outer(eta_array[m], tt[m], grid=False)

    # Inner masks.
    
    """
    if (rr - Rp) <= -x_match*l:
        if t1 >= (tf_inner + t0_inner):
            if eta1 > extr_left and eta1 < extr_right:
    """
    m = np.logical_and(inner_mask,
        np.logical_and(~before_N_wave_mask,
        np.abs( eta_array - cw * np.sign(rr - Rp) * eta_tilde_inner) >= np.sqrt(2 * C_inner * np.abs(tt - t0_inner))
        ))
    Chi[m] = 0.0

    """
    if (rr - Rp) <= -x_match*l:
        if t1 >= (tf_inner + t0_inner):
            if eta1 <= extr_left or eta1 >= extr_right:
    """
    m = np.logical_and(inner_mask,
        np.logical_and(~before_N_wave_mask,
        np.abs( eta_array - cw * np.sign(rr - Rp) * eta_tilde_inner) < np.sqrt(2 * C_inner * np.abs(tt - t0_inner))
        ))
    Chi[m] = (-cw * np.sign(rr[m] - Rp) * eta_array[m] + eta_tilde_inner) / (tt[m] - t0_inner)

    """
    if (rr - Rp) <= -x_match*l:
        if t1 < (tf_inner + t0_inner):
            if eta1 <= eta_inner[-1] and eta1 >= eta_inner[0]:
    """
    interp_inner = RectBivariateSpline(eta_inner, t0_inner + time_inner, solution_inner, kx=1, ky=1)

    m = np.logical_and(inner_mask, 
        np.logical_and(before_N_wave_mask,
        np.logical_and(eta_inner[0] < eta_array,
            eta_array < eta_inner[-1]
        )))
    Chi[m] = -interp_inner(eta_array[m], tt[m], grid=False)


    return Chi

# get the density and velocity perturbations at the grid point from chi
def _get_dens_vel(rr, Chi, gamma, Rp, cw, csp, hr, q, p):

    g1  = _g(rr, Rp, hr, q, p)
    dnl = Chi * 2 / (g1 * (gamma + 1))     # Eq. (11) Bollati et al. 2021

    # Lfu = _Lambda_fu(rr, Rp, csp, hr, gamma, q, p)
    # Lfv = _Lambda_fv(rr, Rp, csp, hr, gamma, q, p)
    # unl = np.sign(rr - Rp) * Lfu * Chi           # Eq. (23) Bollati et al. 2021
    # vnl = np.sign(rr - Rp) * Lfv * Chi * (-cw) # Eq. (24) Bollati et al. 2021 (the sign of v is reversed if we change cw)

    #psi = (np.power(dnl + 1, (gamma-1)/2) - 1) * (gamma+1) / (gamma-1)
    psi = ((gamma+1) / (gamma-1)) * np.sign(dnl + 1) * ((np.abs(dnl + 1)) ** ((gamma - 1) / 2) - 1)

    # get constants
    dOmega_r = np.abs(csp * Rp**-1 * hr**-1 * ((rr / Rp)**(-3 / 2) - 1)) * rr
    c0 = csp * (rr / Rp)**(-q)
    
    # get perturbations
    unl = np.sign(rr - Rp) * (2 * c0) / (gamma+1) * psi
    vnl = (-cw) * c0 * unl / dOmega_r

    return dnl, unl, vnl

# plot the t coordinate as a function of radius
def _plot_r_t(params):
    r = np.linspace(params.r_planet, params.r_outer, 1000)
    times = []
    for rad in r:
        times.append(_t(rad, params.r_planet, params.hr_planet, params.q, params.p))

    plt.plot(r, times)
    plt.xlabel("r")
    plt.ylabel('t')
    plt.show()


# grid.py
# Written by Thomas Hilder

"""
Contains the Grid class on which all Wakeflow models are run/stored.
"""

import numpy                as np
import matplotlib.pyplot    as plt
import astropy.io.fits      as fits
import shutil               as sh
from scipy.interpolate  import RectBivariateSpline
from matplotlib.colors  import LogNorm
from .mcfost_interface  import _read_mcfost_grid_data
from .model_setup       import _Parameters
from .linear_perts      import _LinearPerts
from .non_linear_perts  import _NonLinearPerts

# NOTE: contents are intended mostly for internal use and should not really be accessed by users, but I have still provided
# documentation in the case that advanced users want to mess around with any of it.

# a class to store numpy arrays of results together, so that nothing gets confused. Also stores information and parameters
class _Grid:
    """
    Grid object to generate and store Wakeflow results on. 

    Note that all 3D arrays in the Grid have dimensions (x,z,y) or (phi,z,r).

    Attributes
    ----------
    p : Parameters
        Parameters object (model_setup.py) containing the options used for the model.
    info : dict
        Contains the type, size and contents of the grid.
    height: ndarray
        Scale height of the disk.
    v_r : ndarray
        Contains radial velocities.
    v_phi : ndarray
        Contains azimuthal velocities, positive is counterclockwise.
    rho : ndarray
        Contains densities.
    x : ndarray
        1D. Contains x coordinates of Cartesian grid.
    y : ndarray
        1D. Contains y coordinates of Cartesian grid.
    z_xy : ndarray
        1D. Contains z coordinates of Cartesian grid.
    X : ndarray
        3D. Contains x coordinates of Cartesian grid points.
    Y : ndarray
        3D. Contains y coordinates of Cartesian grid points.
    Z_xy : ndarray
        3D. Contains z coordinates of Cartesian grid points.
    R_xy : ndarray
        3D. Contains radial coordinates of Cartesian grid points.
    PHI_xy : ndarray
        3D. Contains azimuthal coordinates of Cartesian grid points.
    r : ndarray
        1D. Contains radial coordinates of Cylindrical or Mcfost grid.
    phi : ndarray
        1D. Contains azimuthal coordinates of Cylindrical or Mcfost grid.
    z : ndarray
        1D. Contains z coordinates of Cylindrical grid.
    R : ndarray
        3D. Contains radial coordinates of Cylindrical or Mcfost grid points.
    PHI : ndarray
        3D. Contains azimuthal coordinates of Cylindrical or Mcfost grid points.
    Z : ndarray
        3D. Contains z coordinates of Cylindrical or Mcfost grid points.
    """

    # instantiate the class
    def __init__(self, parameters : _Parameters) -> None:
        """
        Parameters
        ----------
        parameters : _Parameters
            _Parameters object (model_setup.py) containing parameters to be used for the Grid.
        """

        # grab parameters object
        self.p = parameters

        # initialise grid properties
        self.info = {
            "Type": None,
            "Log_r": None,
            "Size": [0, 0, 0],
            "Contains": "Empty"
            }

        # initialise arrays
        self.v_r   = None 
        self.v_phi = None 
        self.rho   = None

    # setup the grids with chosen geometry
    def _make_grid(self) -> None:
        """Generates grid according to parameters in self.p and define disk scale height
        """

        # define disk height (not used for mcfost grid)
        self.height = self.p.hr * (self.p.r_outer / self.p.r_ref)**(0.5 - self.p.q) * self.p.r_outer

        # make cartesian grid
        if self.p.grid_type == "cartesian":
            self._make_cartesian_grid()

        # make cylindrical grid
        elif self.p.grid_type == "cylindrical":
            self._make_cylindrical_grid()

        # make mcfost grid
        elif self.p.grid_type == "mcfost":
            self._make_mcfost_grid()

    # setup Cartesian grid
    def _make_cartesian_grid(self) -> None:
        """Generates a Cartesian grid using parameters in self.p
        """
        
        # make grid from specified parameters
        self.x    = np.linspace(-self.p.r_outer, self.p.r_outer, self.p.n_x)
        self.y    = np.linspace(-self.p.r_outer, self.p.r_outer, self.p.n_y)
        self.z_xy = np.linspace(0, self.height, self.p.n_z)

        self.X, self.Z_xy, self.Y  = np.meshgrid(self.x, self.z_xy, self.y, indexing='ij')

        # update grid info
        self.info["Type"]    = "cartesian"
        self.info["Size"][0] = self.x.shape   [0]
        self.info["Size"][1] = self.z_xy.shape[0]
        self.info["Size"][2] = self.y.shape   [0]

    # setup Cylindrical grid
    def _make_cylindrical_grid(self) -> None:
        """Generates a Cylindrical grid using parameters in self.p
        """
         
        # make grid from specifications in parameter file
        if self.p.r_log:
            self.r = np.geomspace(self.p.r_inner, self.p.r_outer, self.p.n_r)
        else:
            self.r = np.linspace (self.p.r_inner, self.p.r_outer, self.p.n_r)    
        #self.phi = np.linspace(0, 2*np.pi, self.p.n_phi)
        self.phi = np.linspace(-np.pi, np.pi, self.p.n_phi)
        self.z   = np.linspace(0, self.height, self.p.n_z)

        self.PHI, self.Z, self.R = np.meshgrid(self.phi, self.z, self.r, indexing='ij')

        # update grid info
        self.info["Type"]    = "cylindrical"
        self.info["Size"][0] = self.phi.shape[0]
        self.info["Size"][1] = self.z.shape  [0]
        self.info["Size"][2] = self.r.shape  [0]
        self.info["Log_r"]   = self.p.r_log

    # setup Mcfost grid by calling Mcfost to generate the geometry, and then read in that geometry
    def _make_mcfost_grid(self) -> None:
        """Generates an Mcfost grid using parameters in self.p
        """

        # generate mcfost grid data from specifications in parameter file
        self.R, self.Z = _read_mcfost_grid_data(self.p)
        self.r         = self.R[0, 0, :]
        self.phi       = np.linspace(0, 2*np.pi, self.p.n_phi)

        PHI_temp = np.repeat(self.phi[:, np.newaxis],    self.p.n_r, axis=1)
        self.PHI = np.repeat(PHI_temp[:, np.newaxis, :], self.p.n_z, axis=1)

        # note that we do not have a self.z because it depends on radius, as the grid is flared

        # update grid info
        self.info["Type"]    = "mcfost"
        self.info["Size"][0] = self.phi.shape[0]
        self.info["Size"][1] = self.Z.shape  [1]
        self.info["Size"][2] = self.r.shape  [0]
        self.info["Log_r"]   = "mcfost"

    # we need (r,phi) points for the Cartesian grid points in order to do transformations
    def _get_r_phi_coords(self) -> None:
        """Find the (r,phi) coordinates corresponding to each (x,y) point
        """

        if self.info["Type"] != "cartesian":
            raise Exception("Can only be used on Cartesian grid.")

        # need the equivalent (r,phi) coordinates when using cartesian grid
        self.R_xy   = np.sqrt   (self.X**2 + self.Y**2)
        self.PHI_xy = np.arctan2(self.Y,     self.X)

    # if the user desires dimensionless results, needs to be used last after all grids are merged
    def _remove_dimensions(self, scale_dens : bool = False) -> None:
        """Convert results to be dimensionless by appropriate scaling; should be used last

        Parameters
        ----------
        scale_dens : bool, optional
            If true, the density is scaled by the rho_ref value in the parameters. Should be left as default for
            scaling perturbations (delta rho / rho) and set to True otherwise.
        """

        grid_length = self.p.U_len / self.p.au

        if self.p.grid_type == "cartesian":
            self.x      /= grid_length
            self.y      /= grid_length
            self.z_xy   /= grid_length
            self.X      /= grid_length
            self.Y      /= grid_length
            self.Z_xy   /= grid_length

        elif self.p.grid_type == "cylindrical":
            self.r      /= grid_length
            self.phi    /= 1.0
            self.z      /= grid_length
            self.R      /= grid_length
            self.PHI    /= 1.0
            self.Z      /= grid_length

        else:
            self.r      /= grid_length
            self.phi    /= 1.0
            self.R      /= grid_length
            self.PHI    /= 1.0
            self.Z      /= grid_length
            
        self.v_r   /= self.p.U_vel
        self.v_phi /= self.p.U_vel

        if scale_dens:
            self.rho /= self.p.rho_ref

    # results for clockwise rotation are the same under a reflection about the y-axis, or phi-axis, 
    # as well as a reverse of sign in v_phi
    def _flip_results(self) -> None:
        """Go from counterclockwise to clockwise rotation; should be used last
        """

        if self.p.grid_type == "cartesian":
            self.y *= -1
            self.Y *= -1
        
        else:
            self.phi *= -1
            self.PHI *= -1

        self.v_phi *= -1

    # setup the unperturbed background disk onto which the perturbations will be mapped
    def _make_keplerian_disk(self) -> None:
        """Generate unperturbed accretion disk in vertical hydrostatic equilibrium, according to parameters in self.p
        """

        # get radii
        if self.info["Type"] == "cartesian":
            self._get_r_phi_coords()
            r = self.R_xy
            z = self.Z_xy
        else:
            r = self.R
            z = self.Z

        # rename for readability in long equations
        p       = self.p.dens_p
        q       = self.p.q
        hr      = self.p.hr
        r_ref   = self.p.r_ref
        r_c     = self.p.r_c
        rho_ref = self.p.rho_ref
        c_s_0   = self.p.c_s_0
        G       = self.p.G_const
        m_star  = self.p.m_star
        m_sol   = self.p.m_solar
        au      = self.p.au

        # Keplerian velocities
        self.v_r   = np.zeros((self.info["Size"][0], self.info["Size"][1], self.info["Size"][2]))
        self.v_kep = 1e-5 * np.sqrt(G * m_star * m_sol / (r * au))
        self.v_phi = np.copy(self.v_kep) * self.p.a_cw

        # get h/r(r)
        hrf = hr * (r / r_ref) ** (0.5 - q)

        # pressure and gravity height dependent correction
        corr = np.sqrt(
            (-1* (p + 2 * q) * hrf**2) + (1 - 2 * q) + (2 * q * r / np.sqrt(r**2 + z**2))
        )

        # perform correction
        self.v_phi *= corr

        # sound speed profile
        c_s = c_s_0 * (r / r_ref)**-q

        # unperturbed density profile (full) (no taper)
        if r_c == 0: 
            self.rho = (
                rho_ref * (r / r_ref)**-p * np.exp(
                    (G * m_star * m_sol / c_s**2) * (1 / np.sqrt((r * au)**2 + (z * au)**2) - 1 / (r * au))
                )
            )
        else: # with taper
            self.rho = (
                rho_ref * (r / r_ref)**-p * np.exp(-(r / r_c)**(2 - p)) * np.exp(
                    (G * m_star * m_sol / c_s**2) * (1 / np.sqrt((r * au)**2 + (z * au)**2) - 1 / (r * au))
                )
            )

        # unperturbed density profile (thin disk approx.)
        #self.rho = rho_ref * (r / r_ref)**-p * np.exp( - 0.5 * (z / r_ref)**2 * hr**-2 * (r / r_ref)**(2 * q - 3))

        # update grid info
        self.info["Contains"] = "Keplerian velocity field, unperturbed density"

    # fill the grids with zeros, used for storing perturbations on separate grids before mapping onto background
    def _make_empty_disk(self) -> None:
        """fill v_r, v_phi and rho with zeros
        """

        # make grids full of zeros
        self.v_r   = np.zeros((self.info["Size"][0], self.info["Size"][1], self.info["Size"][2]))
        self.v_phi = np.zeros((self.info["Size"][0], self.info["Size"][1], self.info["Size"][2]))
        self.rho   = np.zeros((self.info["Size"][0], self.info["Size"][1], self.info["Size"][2]))

        # update grid info
        self.info["Contains"] = "Zeros"

    # add results from the linear regime nearby the planet onto the grid
    def _add_linear_perturbations(self, LinearPerts : _LinearPerts, rho_background : "_Grid.rho") -> None:
        """Add results from the linear regime, stored in LinearPerts object, to the grid

        Parameters
        ----------
        LinearPerts : LinearPerts
            LinearPerts object containing results from the linear regime.
        rho_background : Grid.rho
            unperturbed density from Grid object where make_keplerian_disk has been used.
        """

        # box size (in units of Hill radius), note for conversions that self.p.l = 1 Hill radius in cgs
        x_box_size = 2 * self.p.scale_box
        y_box_size = 2 * self.p.scale_box_ang

        min_r = self.p.r_planet - x_box_size * self.p.l
        max_r = self.p.r_planet + x_box_size * self.p.l

        min_y = -y_box_size * self.p.l
        max_y =  y_box_size * self.p.l

        max_phi =  np.pi / 2
        min_phi = -np.pi / 2

        # find (phi, r) grid for either Cartesian or Cylindrical global grid
        if self.info["Type"] == "cartesian":
            self._get_r_phi_coords()
            R, PHI = self.R_xy, self.PHI_xy
        else:
            R, PHI = self.R, self.PHI

        # new PHI grid to use (-pi,pi) instead of (0,2pi), where values are swapped in place, also ditch z coordinate
        # also construct a mask that contains 0 outside linear annulus and 1 inside
        PHI_new     = np.zeros((PHI.shape[0],PHI.shape[2]))
        Y_new       = np.zeros((PHI.shape[0],PHI.shape[2]))
        linear_mask = np.zeros((PHI.shape[0],PHI.shape[2]))

        R_new = R[:,0,:]

        for i in range(PHI.shape[0]):
            for j in range(PHI.shape[2]):

                # transforming phi coordinate in place
                if PHI[i,0,j] > np.pi:
                    PHI_new[i,j] = PHI[i,0,j] - 2*np.pi
                else:
                    PHI_new[i,j] = PHI[i,0,j]

                Y_new[i,j] = R[i,0,j] * np.sin(PHI[i,0,j])

                # constructing mask
                if PHI_new[i,j] > min_phi and PHI_new[i,j] < max_phi \
                    and Y_new[i,j] > min_y and Y_new[i,j] < max_y \
                    and R_new[i,j] > min_r and R_new[i,j] < max_r:
                    linear_mask[i,j] = 1
                else:
                    linear_mask[i,j] = 0

        # get linear solution           
        lp = LinearPerts

        # assemble interpolation functions over linear perts grid
        interp_v_r   = RectBivariateSpline(lp.y_ann, lp.r_ann, lp.pert_v_r_ann)
        interp_v_phi = RectBivariateSpline(lp.y_ann, lp.r_ann, lp.pert_v_phi_ann)
        interp_rho   = RectBivariateSpline(lp.y_ann, lp.r_ann, lp.pert_rho_ann)

        # evaluate interpolations on global grid
        global_lin_v_r   = interp_v_r.ev  (Y_new, R_new)
        global_lin_v_phi = interp_v_phi.ev(Y_new, R_new)
        global_lin_rho   = interp_rho.ev  (Y_new, R_new)

        # apply mask to only get solution in valid domain
        global_lin_v_r   = global_lin_v_r   * linear_mask
        global_lin_v_phi = global_lin_v_phi * linear_mask
        global_lin_rho   = global_lin_rho   * linear_mask

        # Add velocities, identical at all heights for now
        self.v_r   += global_lin_v_r  [:, np.newaxis, :]
        self.v_phi += global_lin_v_phi[:, np.newaxis, :]

        # Add density, scaling by background density
        self.rho += global_lin_rho[:, np.newaxis, :] * rho_background

        # plot for debugging
        #_, ax = plt.subplots(subplot_kw=dict(projection='polar'))
        #myplot = ax.contourf(PHI[:,0,:], R[:,0,:], global_lin_v_r, levels=300, cmap='RdBu')
        #ax.set_ylim(0, self.p.r_outer)
        #plt.colorbar(myplot)
        #plt.show()

        # update grid info
        self.info["Contains"] = "linear perturbations"

    # add results from non-linear wake propagation onto the grid
    def _add_non_linear_perturbations(self, NonLinearPerts : _NonLinearPerts, rho_background : "_Grid.rho") -> None:
        """Add results from the non-linear regime, stored in NonLinearPerts object, to the grid

        Parameters
        ----------
        NonLinearPerts : NonLinearPerts
            NonLinearPerts object containing results from the non-linear regime.
        rho_background : Grid.rho
            unperturbed density from Grid object where make_keplerian_disk has been used.
        """

        nonlin = NonLinearPerts

        # Add velocities, identical at all heights for now
        nl_vr   = nonlin.vr  [:, np.newaxis, :]
        nl_vphi = nonlin.vphi[:, np.newaxis, :]

        self.v_r   += nl_vr
        self.v_phi += nl_vphi

        # Define scale height array
        if self.info["Type"] == "cartesian":

            self._get_r_phi_coords()
            r = self.R_xy
            z = self.Z_xy

            size_ones = (len(self.x), self.p.n_z, len(self.y))

        else:

            r = self.R 
            z = self.Z

            size_ones = (self.p.n_phi, self.p.n_z, self.p.n_r)

        self.H = self.p.h_ref * (r / self.p.r_ref)**self.p.beta

        # Add density, scaling vertically as per thin disk assumption
        #nl_rho = (np.ones(size_ones) + nonlin.rho[:, np.newaxis, :]) * self.p.rho_ref * (r/self.p.r_ref)**(-self.p.p)
        #self.rho = nl_rho * np.exp(-0.5 * (z / self.H)**2)

        self.rho += nonlin.rho[:, np.newaxis, :] * rho_background

        # update grid info
        self.info["Contains"] = "non-linear perturbations"

#    def add_phantom_midplane(self, PhantomDump):
#
#        PD = PhantomDump
#
#        # Add velocities, identical at all heights
#        ph_vr   = PD.vr  [:, np.newaxis, :]
#        ph_vphi = PD.vphi[:, np.newaxis, :]
#
#        self.v_r   += ph_vr #/(1 + self.Z)
#        self.v_phi += ph_vphi #/(1 + self.Z)
#
#        # Define scale height array
#        self.H = self.p.h_ref * (self.R / self.p.r_ref)**self.p.beta
#
#        # Add density, scaling vertically as per thin disk assumption
#        ph_rho   = PD.rho[:, np.newaxis, :]
#        self.rho = ph_rho * np.exp(-0.5 * (self.Z / self.H)**2)
#
#        # update grid info
#        self.info["Contains"] = "phantom mid-plane extrapolated to 3D"

    # take multiple grid objects and merge them if the geometry is the same, simply by adding the components
    def _merge_grids(self, grid_to_merge : "_Grid") -> None:
        """Merge v_r, v_phi and rho from two compatible Grid objects (ie. with the same _Parameters)
        """

        g = grid_to_merge

        if type(grid_to_merge) is not _Grid:
            raise Exception("merge_grids must be given a Grid object. Merging failed.")
        
        if self.p != g.p:
            raise Exception("merge_grids requires both grids must have same parameters. Merging failed.")

        # merge data arrays
        self.v_r   += g.v_r
        self.v_phi += g.v_phi
        self.rho   += g.rho

        # update info
        self.info["Contains"] += " AND " + g.info["Contains"]

#    def merge_phantom_densities(self, grid_to_merge):
#
#        g = grid_to_merge
#
#        if type(grid_to_merge) is not Grid:
#            print("Must be given a Grid object")
#            return False
#        
#        if self.p != g.p:
#            print("Both grids must have same type parameters")
#            return False
#
#        # merge data arrays
#        self.rho = g.rho

    # create plots of a constant z slice, mostly used for debugging but also shows midplane results to user
    # if they set show_midplane_plots=True
    def _show_disk2D(
        self, 
        z_slice :    int, 
        show :      bool = False, 
        save :      bool = False, 
        dimless :   bool = False, 
        vphi_lim : float = 1.0, 
        cw :        bool = False
    ) -> None:
        """Generate plots of the disk at constant z value with index z_slice

        For example, mid-plane plots are generated with z_slice=0.

        Parameters
        ----------
        z_slice : int
            Index in Grid of z_slice to plot.
        show : bool, optional
            True to show plots to user using matplotlib.pyplot.show(). Default is False.
        save : bool, optional
            True to save plots in directory with results. Default is False.
        dimless: bool, optional
            Set to True if using dimensionless results to get accurately labelled plots.
        vphi_lim : float, optional
            Modifier for the limits on the v_phi plot.
        cw : bool, optional
            True for clockwise rotation and False for anticlockwise. Since anticlockwise is default, False is default.
        """

        # parameters
        savedir      = f"{self.p.system}/{self.p.name}/{self.p.m_planet}Mj"
        contour_lvls = 300

        # find maximum contours
        vr_max   = np.max(self.v_r  [:,z_slice,:])
        vphi_max = np.max(self.v_phi[:,z_slice,:]) * vphi_lim
        if cw == True:
            vphi_max = np.min(self.v_phi[:,z_slice,:]) * vphi_lim
        rho_max  = np.percentile(self.rho[:,z_slice,:], 99.9)

        # Cartesian grid plots
        if self.info["Type"] == "cartesian":

            # v_r
            plt.close("all")
            fig, ax = plt.subplots(dpi=150)
            c       = ax.pcolormesh(self.x, self.y, np.transpose(self.v_r[:,z_slice,:]), vmin=-vr_max, vmax=vr_max, cmap='RdBu', rasterized=True)
            ax.axis('scaled')
            ax.set_title(r"$\delta v_r$")
            if not dimless:
                ax.set_xlabel(r"$x \, [\mathrm{AU}]$")
                ax.set_ylabel(r"$y \, [\mathrm{AU}]$")
                fig.colorbar(c, extend="both", label="$[\mathrm{km / s}]$")
            else:
                ax.set_xlabel(r"$x$")
                ax.set_ylabel(r"$y$")
                fig.colorbar(c, extend="both")
            if save:
                plt.savefig(f'{savedir}/vr_z{z_slice}.pdf')
            if show:
                plt.show()

            # v_phi
            plt.close("all")
            fig, ax = plt.subplots(dpi=150)
            c       = ax.pcolormesh(self.x, self.y, np.transpose(self.v_phi[:,z_slice,:]), vmin=-vphi_max, vmax=vphi_max, cmap='RdBu', rasterized=True)
            ax.axis('scaled')
            ax.set_title(r"$\delta v_{\phi}$")
            if not dimless:
                ax.set_xlabel(r"$x \, [\mathrm{AU}]$")
                ax.set_ylabel(r"$y \, [\mathrm{AU}]$")
                fig.colorbar(c, extend="both", label="$[\mathrm{km / s}]$")
            else:
                ax.set_xlabel(r"$x$")
                ax.set_ylabel(r"$y$")
                fig.colorbar(c, extend="both")
            if save:
                plt.savefig(f'{savedir}/vphi_z{z_slice}.pdf')
            if show:
                plt.show()

            # rho
            plt.close("all")
            fig, ax = plt.subplots(dpi=150)
            c       = ax.pcolormesh(self.x, self.y, np.transpose(self.rho[:,z_slice,:]), vmin=-rho_max, vmax=rho_max, cmap='RdBu', rasterized=True)
            ax.axis('scaled')
            ax.set_title(r"$\delta \rho \, / \rho$")
            if not dimless:
                ax.set_xlabel(r"$x \, [\mathrm{AU}]$")
                ax.set_ylabel(r"$y \, [\mathrm{AU}]$")
            else:
                ax.set_xlabel(r"$x$")
                ax.set_ylabel(r"$y$")
            fig.colorbar(c, extend="both")
            if save:
                plt.savefig(f'{savedir}/rho_z{z_slice}.pdf')
            if show:
                plt.show()

        # Polar grid plots
        else:

            # v_r
            plt.close("all")
            fig, ax = plt.subplots(dpi=150)
            c       = ax.pcolormesh(self.phi, self.r, np.transpose(self.v_r[:,z_slice,:]), vmin=-vr_max, vmax=vr_max, cmap='RdBu', rasterized=True)
            ax.set_title(r"$\delta v_r$")
            if not dimless:
                ax.set_xlabel(r"$\phi \, [\mathrm{rad}]$")
                ax.set_ylabel(r"$r \, \, [\mathrm{AU}]$")
                fig.colorbar(c, extend="both", label="$[\mathrm{km / s}]$")
            else:
                ax.set_xlabel(r"$\phi \, [\mathrm{rad}]$")
                ax.set_ylabel(r"$r$")
                fig.colorbar(c, extend="both")
            if save:
                plt.savefig(f'{savedir}/vr_z{z_slice}.pdf')
            if show:
                plt.show()

            # v_phi
            plt.close("all")
            fig, ax = plt.subplots(dpi=150)
            c       = ax.pcolormesh(self.phi, self.r, np.transpose(self.v_phi[:,z_slice,:]), vmin=-vphi_max, vmax=vphi_max, cmap='RdBu', rasterized=True)
            ax.set_title(r"$\delta v_\phi$")
            if not dimless:
                ax.set_xlabel(r"$\phi \, [\mathrm{rad}]$")
                ax.set_ylabel(r"$r \, \, [\mathrm{AU}]$")
                fig.colorbar(c, extend="both", label="$[\mathrm{km / s}]$")
            else:
                ax.set_xlabel(r"$\phi \, [\mathrm{rad}]$")
                ax.set_ylabel(r"$r$")
                fig.colorbar(c, extend="both")
            if save:
                plt.savefig(f'{savedir}/vphi_z{z_slice}.pdf')
            if show:
                plt.show()

            # rho
            plt.close("all")
            fig, ax = plt.subplots(dpi=150)
            c       = ax.pcolormesh(self.phi, self.r, np.transpose(self.rho[:,z_slice,:]), vmin=-rho_max, vmax=rho_max, cmap='RdBu', rasterized=True)
            ax.set_title(r"$\delta \rho \, / \rho$")
            if not dimless:
                ax.set_xlabel(r"$\phi \, [\mathrm{rad}]$")
                ax.set_ylabel(r"$r \, \, [\mathrm{AU}]$")
            else:
                ax.set_xlabel(r"$\phi \, [\mathrm{rad}]$")
                ax.set_ylabel(r"$r$")
            fig.colorbar(c, extend="both")
            if save:
                plt.savefig(f'{savedir}/rho_z{z_slice}.pdf')
            if show:
                plt.show()
        
        plt.close("all")
    
    # create a fits file from the results formatted specifically for use in MCFOST in combination with
    # the .para file written by pymcfost through wakeflow
    def _write_fits_file(self) -> None:
        """Write a .FITS file from the results compatible with being run in MCFOST
        """

        if self.info["Type"] == "cartesian":

            print("WARNING: Cannot write FITS file when using Cartesian grid.")

        else:

            # create empty array for vertical velocities
            self.v_z = np.zeros((self.p.n_phi, self.p.n_z, self.p.n_r))

            # create master velocities array
            velocities = 1e3 * np.array([self.v_r, self.v_phi, self.v_z])

            # setup HDUs for fits file
            primary_hdu  = fits.PrimaryHDU(np.abs(self.rho))
            second_hdu   = fits.ImageHDU  (np.abs(self.rho)) # * np.exp(-1 * (self.Z / self.H)**2)
            tertiary_hdu = fits.ImageHDU  (velocities)

            # set header properties for mcfost
            primary_hdu.header['hierarch read_gas_velocity'] = 2
            primary_hdu.header['hierarch gas_dust_ratio']    = 100
            primary_hdu.header['hierarch read_gas_density']  = 1
            primary_hdu.header['read_n_a']                   = 0

            # sink particles properties in header (code modifications needed for anything other than 2)
            primary_hdu.header['hierarch N_sink'] = 2

            # star properties in header
            primary_hdu.header['hierarch M_star']  = self.p.m_star       # star mass in M_sol
            primary_hdu.header['hierarch x_star']  = 0.                  # star at origin
            primary_hdu.header['hierarch y_star']  = 0.
            primary_hdu.header['hierarch z_star']  = 0.
            primary_hdu.header['hierarch vx_star'] = 0.                  # star not moving
            primary_hdu.header['hierarch vy_star'] = 0.
            primary_hdu.header['hierarch vz_star'] = 0.

            # planet properties in header
            primary_hdu.header['hierarch M_planet']  = self.p.m_planet    # planet mass in M_jup
            primary_hdu.header['hierarch x_planet']  = self.p.r_planet    # planet x position simply orbital radius [AU]
            primary_hdu.header['hierarch y_planet']  = 0.
            primary_hdu.header['hierarch z_planet']  = 0.
            primary_hdu.header['hierarch vx_planet'] = 0.                # planet velocity just keplerian rotation [m/s] (positive for anticlockwise, which is cw=-1)
            primary_hdu.header['hierarch vy_planet'] = np.sqrt(
                self.p.G_const*self.p.m_star*self.p.m_solar/(self.p.r_planet*self.p.au)
                ) / 100 * self.p.a_cw
            primary_hdu.header['hierarch vz_planet'] = 0.

            # setup fits file
            hdul = fits.HDUList([primary_hdu, second_hdu, tertiary_hdu])

            # Write a fits file for mcfost
            fitsname = "wakeflow_model.fits"
            hdul.writeto(f"{self.p.system}/{self.p.name}/{self.p.m_planet}Mj/{fitsname}", overwrite=True)
            print("Saved FITS file formatted for MCFOST.")

            # Copy mcfost parameter file here too
            sh.copy(
                f"{self.p.system}/{self.p.name}/mcfost/mcfost_{self.p.name}.para",
                f"{self.p.system}/{self.p.name}/{self.p.m_planet}Mj/mcfost.para"
            )
    
    # simply saves the grid and densities/velocities to .npy files so that they may be read later
    def _save_results(self, label : str, printed : str) -> None:
        """ Save results and grid to .npy files in results directory

        Parameters
        ----------
        label : str
            Prepended to saved file names (used in wakeflow.py to give "delta" for perturbations and "total" for perturbations 
            applied on top of an unperturbed disk).
        printed : str
            Message to be printed to the user in format f"{printed} saved to {savedir}".
        """

        savedir = f"{self.p.system}/{self.p.name}/{self.p.m_planet}Mj"

        # save grid:
        if self.p.grid_type == "cartesian":
            np.save(f"{savedir}/X.npy", self.X)
            np.save(f"{savedir}/Z.npy", self.Z_xy)
            np.save(f"{savedir}/Y.npy", self.Y)
        else:
            np.save(f"{savedir}/{label}_PHI.npy", self.PHI)
            np.save(f"{savedir}/{label}_Z.npy", self.Z)
            np.save(f"{savedir}/{label}_R.npy", self.R)

        # save results:
        np.save(f"{savedir}/{label}_v_r.npy", self.v_r)
        np.save(f"{savedir}/{label}_v_phi.npy", self.v_phi)
        np.save(f"{savedir}/{label}_rho.npy", self.rho)

        print(f"{printed} saved to {savedir}")

        


# wakefit_interface.py
# Written by Thomas Hilder

"""
Contains the WakefitModel class, intended for use in Wakefit to generate wake models
"""

import subprocess, os
from .model_setup         import _run_setup, _Parameters
from .grid                import _Grid
from .linear_perts        import _LinearPerts
from .non_linear_perts    import _NonLinearPerts
from .wakeflow            import WakeflowModel

# class for wakefit to use to generate velocity perturbations for models
class _WakefitModel(WakeflowModel):

    # generate the model using the configuration specified by the user
    def run(self) -> None:

        # check requirements for models for wakefit are satisfied (don't make files!)
        assert self.model_params["make_midplane_plots"] == False
        assert self.model_params["save_total"]          == False
        assert self.model_params["save_perturbations"]  == False
        
        # run setup
        try:
            params = _run_setup(self.model_params)
        except ArithmeticError:
            raise Exception("Model has not been configured.")

        # run wakeflow using custom method for wakefit
        return self._run_wakeflow_for_wakefit(params)

    # internal method that is called by self.run to generate the results for a specific set of parameters
    def _run_wakeflow_for_wakefit(self, params: _Parameters) -> None:
        """
        Internal use method for generating the planet wake by calling other parts of Wakeflow.

        Parameters
        ----------
        params : Parameters 
            Paramaters object specfiying options for the model to be generated.
        """

        print("Extracting linear perturbations nearby planet")

        # make empty grid for linear perturbations
        grid_lin_perts = _Grid(params)
        grid_lin_perts._make_grid()
        grid_lin_perts._make_empty_disk()

        # extract linear perturbations from file
        lin_perts = _LinearPerts(params)
        lin_perts._cut_box_annulus_segment()

        # add the linear perturbations onto grid
        grid_lin_perts._add_linear_perturbations(lin_perts, rho_background=0.0)

        # make empty grid for non-linear perturbations
        grid_nonlin_perts = _Grid(params)
        grid_nonlin_perts._make_grid()
        grid_nonlin_perts._make_empty_disk()

        # initialise non-linear perturbations
        nonlin_perts = _NonLinearPerts(params, grid_nonlin_perts)

        # extract initial condition from the linear perturbations
        nonlin_perts._extract_ICs(lin_perts)

        # solve for non-linear perturbations
        nonlin_perts._get_non_linear_perts()

        # add non-linear perturbations to grid
        grid_nonlin_perts._add_non_linear_perturbations(nonlin_perts, rho_background=0.0)

        # merge grids for result
        grid_lin_perts._merge_grids(grid_nonlin_perts)

        # return the grid and velocity components
        return (grid_lin_perts.x, grid_lin_perts.y, grid_lin_perts.v_r, grid_lin_perts.v_phi)


# burgers.py
# Written by Thomas Hilder and Francesco Bollati

"""
Contains the solve_burgers function used in the wake non-linear propagation calculations.
"""

import numpy                as np
import matplotlib.pyplot    as plt

# NOTE: contents are intended for internal use and should not be directly accessed by users

# function to solve burgers equation as part of the non-linear propagation of the planet wake
def _solve_burgers(
    eta, 
    profile, 
    gamma, 
    beta_p, 
    C, 
    CFL, 
    eta_tilde, 
    t0, 
    linear_solution, 
    linear_t, 
    show_teta, 
    tf_fac
): 
    """Propagate the wake in (t,eta,chi) space by solving Eq. 10 from Bollati et al. 2021 using Godunov scheme.
    """

    # Eq. (15) Bollati et al. 2021
    profile = profile * (gamma + 1) * beta_p / 2**(3 / 4)
    deta    = eta[1] - eta[0]

    # time required to develop N-wave for betap = 1
    tf_th = 300 

    # time required to display N-wave for generic betap, Eq. (39) Rafikov 2002
    tf  = tf_th / beta_p 
    tf *= tf_fac

    # Eq. (18) Bollati et al. 2021
    eta_min = -eta_tilde - np.sqrt(2 * C * tf) - 3 
    eta_max = -eta_tilde + np.sqrt(2 * C * tf) + 3

    # extend the profile domain due to profile broadening during evolution 
    extr = eta[-1]

    while extr < eta_max:

        eta     = np.append(eta, eta[-1] + deta)
        profile = np.append(profile, 0)
        extr    = eta[-1]

        #if show_teta:
        #    linear_solution = np.append(profile, 0, axis=0)

    extr = eta[0]

    while extr > eta_min:

        eta     = np.insert(eta, 0, eta[0] - deta)
        profile = np.insert(profile, 0, 0)
        extr    = eta[0]

        #if show_teta:
        #    linear_solution = np.insert(linear_solution, 0, 0, axis=0)

    # number of centers
    Neta = len(eta)

    a = eta[0] - deta/2

    # cells edges
    x = np.zeros(Neta + 1) 
    for i in range(0, Neta + 1):
        x[i] = a + i * deta

    solution = [np.zeros((Neta), dtype=float)]
    time = [0]

    # linear solution as initial condition
    solution[0] = profile

    # define flux vector
    F = np.zeros(Neta + 1) #there is a flux at each cell edge

    # calculate flux according to Godunov scheme, vectorised
    def _NumericalFluxVector(uL, uR):
        """Calculate flux according to the Godunov scheme.
        """

        FL = 0.5 * uL**2
        FR = 0.5 * uR**2

        # compute the shock speed
        s = 0.5 * (uL + uR)

        # See "GodunovNumericalFlux".
        # This returns the same value without branching
        # through the if statements.
        return  (uL >= uR) * (
            (s > 0.0) * FL + (s <= 0.0) * FR
            ) + (uL < uR) * (
                (uL > 0.0) * (FL) +
                (uL <= 0.0) * (uR < 0.0) * FR
            )

    # initialise
    lapsed_time = 0
    counter     = 0

    # time integration
    while lapsed_time < tf:

        # increment
        counter += 1

        # time step calculation
        dt = min(deta * CFL / (max(abs(solution[-1])) + 1e-8), 0.02)

        # update time
        time.append(time[-1] + dt)
        lapsed_time += dt

        # compute the interior fluxes
        F[1:Neta] = _NumericalFluxVector(solution[-1][0:-1], solution[-1][1:])

        # compute the left boundary flux
        if solution[-1][0] < 0.0:
            uL = 2.0 * solution[-1][0] - solution[-1][1]
        else:
            uL = solution[0][0]

        uR   = solution[-1][0]
        F[0] = _NumericalFluxVector(uL, uR)

        # compute the right boundary flux
        if solution[-1][Neta - 1] > 0.0:
            uR = 2.0 * solution[-1][Neta - 1] - solution[-1][Neta - 2]
        else:
            uR = solution[0][Neta - 1]

        uL      = solution[-1][Neta - 1]
        F[Neta] = _NumericalFluxVector(uL, uR)

        solution.append(solution[-1][0:Neta] - dt / deta * (F[1:Neta+1] - F[0:Neta]))

    solution = np.array(solution).transpose()
    time     = np.array(time)

    # plot Fig. 3 Bollati et al. 2021
    #plt.plot(eta, solution[:,0], label = "$t=t_0+$ ")#+str(0*dt))
    #plt.plot(eta, solution[:,100], label = "$t=t_0+$ ")#+str(round(dt*Nt/10,2)))
    #plt.plot(eta, solution[:,200], label = "$t=t_0+$ ")#+str(round(dt*Nt/6,2)))
    #plt.plot(eta, solution[:,300], label = "$t=t_0+$ ")#+str(round(dt*Nt/3,2)))
    #plt.plot(eta, solution[:,400], label = "$t=t_0+$ ")#+str(round(T,2)))
    #plt.legend()
    #plt.xlabel(r"$\eta$")
    #plt.ylabel(r"$\chi(t,\eta)$")
    #plt.grid(True)
    #plt.title(r'$\chi$ "evolution" $r > r_p$')
    #plt.show()

    #if show_teta: # combining linear and non-linear solution and plotting
    #    
    #    # scale linear solution
    #    linear_solution = linear_solution * (gamma + 1) * beta_p / 2**(3 / 4)
    #
    #    # add linear solution in (t,eta) to non-linear solution array
    #    total_solution = np.concatenate((linear_solution, solution), axis=1)
    #    total_time     = np.concatenate((linear_t,        time + t0))
    #
    #    fig, ax = plt.subplots(1)
    #    cont = ax.contourf(total_time, eta, total_solution, levels=np.arange(-4, 4, 0.05), cmap='RdBu')
    #    for c in cont.collections:
    #        c.set_rasterized(True)
    #    plt.colorbar(cont, label='$\chi$')
    #    ax.set_xlim(0,10)
    #    ax.set_xlabel('$t$')
    #    ax.set_ylabel('$\eta$')
    #    #plt.savefig("teta_badjoin.pdf")
    #    plt.show()

    # The following will put the linear solution into returned solution for Chi
    #if False:
    #    Nt = np.shape(total_solution)[1]
    #
    #    solution_inner = np.zeros(total_solution.shape) # solution for r < Rp (and disc rotating counterclockwise)
    #    for i in range(Neta):
    #        for j in range(Nt):
    #            solution_inner[i,j] = total_solution[int(Neta-1-i),j]
    #
    #    eta_inner = - eta[::-1]
    #    return total_time, eta, total_solution, eta_inner, solution_inner

    return time, eta, solution
    


# non_linear_perts.py
# Written by Thomas Hilder, Daniele Fasano and Francesco Bollati

"""
Contains the NonLinearPerts class responsible for handling the non-linear regime of the models.
"""

# type hinting without circular imports
from typing import          TYPE_CHECKING

import time
import numpy                    as np
import matplotlib.pyplot        as plt
from scipy.interpolate      import griddata
from copy                   import copy
from tqdm                   import tqdm
from .burgers               import _solve_burgers
from .transformations       import _Eta, _t, _get_chi, _get_dens_vel, _plot_r_t, _t_vector, _get_chi_vector

if TYPE_CHECKING:
    from .model_setup       import _Parameters
    from .grid              import _Grid
    from .linear_perts      import _LinearPerts

# NOTE: contents are intended for internal use and should not be directly accessed by users

# class for non-linear regime results
class _NonLinearPerts():
    """
    Class to store the results from the non-linear wake propagation away from the planet.
    """

    # intialise object, give params and grid
    def __init__(self, parameters: '_Parameters', Grid: '_Grid') -> None:
        """Setup _NonLinearPerts object by grabbing _Parameters and _Grid objects.
        """
        
        # grab parameters object
        self.p = parameters

        # should be handed an empty grid with the correct dimensions and grid setup used in the run
        self.g = Grid

    # get initial conditions for burgers eqn
    def _extract_ICs(self, LinearPerts: '_LinearPerts') -> None:
        """Extract the initial conditions for the non-linear evolution from the linear regime.
        """
        
        # grab linear perturbations object
        lp = LinearPerts

        # grid
        x = lp.X[0,:]
        y = lp.Y[:,0]

        # set maximum eta - semi-width of the support of the azimuthal profile of the linear density perturbation
        eta_max = 25 

        # inner and outer wake x position
        x_box_outer =  lp.x_box
        x_box_inner = -lp.x_box

        # find the index in the x grid corresponding to the edge of the box
        index_outer = np.argmin(np.abs(x - x_box_outer))
        index_inner = np.argmin(np.abs(x - x_box_inner))

        # extract profile of constant x along edge of box for initial condition
        profile_outer = lp.pert_rho[:,index_outer] / np.sqrt(np.abs(x_box_outer))
        profile_inner = lp.pert_rho[:,index_inner] / np.sqrt(np.abs(x_box_inner))

        # restrict y range --- I'm not sure why this is necessary but it is. Larger values will cause issues
        y_max = 30
        y_rest = y[(y > -y_max) & (y < y_max)]

        # restrict inner and outer IC
        profile_rest_outer = profile_outer[(y > -y_max) & (y < y_max)]
        profile_rest_inner = profile_inner[(y > -y_max) & (y < y_max)]

        # ## find eta points for IC profile using full transformation
        # find local cart. coords in real units
        x_IC_outer = self.p.l * np.repeat(x_box_outer, len(y_rest))
        y_IC_outer = self.p.l * y_rest
        x_IC_inner = self.p.l * np.repeat(x_box_inner, len(y_rest))
        y_IC_inner = self.p.l * y_rest

        # find corresponding global polar coords
        r_IC_outer   = x_IC_outer + self.p.r_planet
        r_IC_inner   = x_IC_inner + self.p.r_planet
        phi_IC_outer = y_IC_outer / self.p.r_planet
        phi_IC_inner = y_IC_inner / self.p.r_planet

        if False:
            plt.plot(phi_IC_outer, profile_rest_outer, label="outer")
            plt.plot(phi_IC_inner, profile_rest_inner, label="inner")
            plt.legend(loc="best")
            plt.show()

        # initialise arrays for corresponding t,eta points
        eta_IC_outer = np.zeros(len(y_rest))
        eta_IC_inner = np.zeros(len(y_rest))
        t_IC_outer   = np.zeros(len(y_rest))
        t_IC_inner   = np.zeros(len(y_rest))

        # perform transformation
        for i in range(len(y_rest)):

            # eta for outer and inner
            eta_IC_outer[i] = _Eta(r_IC_outer[i], phi_IC_outer[i], self.p.r_planet, self.p.hr_planet, self.p.q, -1)
            eta_IC_inner[i] = _Eta(r_IC_inner[i], phi_IC_inner[i], self.p.r_planet, self.p.hr_planet, self.p.q, -1)

            # t for outer and inner
            t_IC_outer[i] = _t(r_IC_outer[i], self.p.r_planet, self.p.hr_planet, self.p.q, self.p.p)
            t_IC_inner[i] = _t(r_IC_inner[i], self.p.r_planet, self.p.hr_planet, self.p.q, self.p.p)

        # restrict eta range using eta_max
        self.eta_outer = eta_IC_outer[(eta_IC_outer > -eta_max) & (eta_IC_outer < eta_max)]
        self.eta_inner = eta_IC_inner[(eta_IC_inner > -eta_max) & (eta_IC_inner < eta_max)]
        self.profile_outer = profile_rest_outer[(eta_IC_outer > -eta_max) & (eta_IC_outer < eta_max)]
        self.profile_inner = profile_rest_inner[(eta_IC_inner > -eta_max) & (eta_IC_inner < eta_max)]

        # set t0
        self.t0_outer = t_IC_outer[0]
        self.t0_inner = t_IC_inner[0]

        # old, approximate IC extraction procedure
        if False:
            y_match = -np.sign(lp.x_box) * 0.5 * lp.x_box**2
            y_cut = y[(y - y_match > -eta_max) & (y - y_match < eta_max)]
            self.profile = profile[(y - y_match > -eta_max) & (y - y_match < eta_max)]
            self.eta = y_cut - y_match*np.ones(len(y_cut))
            self.t0 = _t(self.p.r_planet + self.p.l * lp.x_box, self.p.r_planet, self.p.hr, self.p.q, self.p.p)

        # set eta_tilde for outer wake:
        for i in range(len(self.eta_outer)):
            if self.profile_outer[i] == 0 and self.eta_outer[i] > -10 and self.eta_outer[i] < 0:
                zero_outer = self.eta_outer[i]
            elif i!= (len(self.eta_outer) - 1) and self.profile_outer[i] * self.profile_outer[i + 1] < 0 and self.eta_outer[i] > -10 and self.eta_outer[i] < 0:
                zero_outer = 0.5 * (self.eta_outer[i] + self.eta_outer[i + 1])
        self.eta_tilde_outer = -zero_outer

        # set eta_tilde for inner wake:
        for i in range(len(self.eta_inner)):
            if self.profile_inner[i] == 0 and self.eta_inner[i] > -10 and self.eta_inner[i] < 0:
                zero_inner = self.eta_inner[i]
            elif i!= (len(self.eta_inner) - 1) and self.profile_inner[i] * self.profile_inner[i + 1] < 0 and self.eta_inner[i] > -10 and self.eta_inner[i] < 0:
                zero_inner = 0.5 * (self.eta_inner[i] + self.eta_inner[i + 1])
        self.eta_tilde_inner = -zero_inner

        # set C for outer wake:
        deta_outer = self.eta_outer[1] - self.eta_outer[0]
        profile0_outer = self.profile_outer[self.eta_outer < -self.eta_tilde_outer]
        C0_outer = -np.trapz(profile0_outer, dx = deta_outer)
        self.C_outer = (self.p.gamma + 1) * (self.p.m_planet / self.p.m_thermal) * C0_outer / 2**(3/4)

        # set C for inner wake:
        deta_inner = self.eta_inner[1] - self.eta_inner[0]
        profile0_inner = self.profile_inner[self.eta_inner < -self.eta_tilde_inner]
        C0_inner = -np.trapz(profile0_inner, dx = deta_inner)
        self.C_inner = (self.p.gamma + 1) * (self.p.m_planet / self.p.m_thermal) * C0_inner / 2**(3/4)

        #print('     Outer Wake:')
        #print('         eta_tilde = ', self.eta_tilde_outer)
        #print('         C0 = ', C0_outer)
        #print('         t0 = ', self.t0_outer)
        #print('     Inner Wake:')
        #print('         eta_tilde = ', self.eta_tilde_inner)
        #print('         C0 = ', C0_inner)
        #print('         t0 = ', self.t0_inner)

        # ======================================
        # === put linear solution into t,eta ===

        # you will need to run cut_box_square on the linear perts object before doing this.
        # it is okay to run it as well as the annulus cut, they don't overwrite each other

        beta_p = self.p.m_planet / self.p.m_thermal
        gamma  = self.p.gamma

#        if self.p.show_teta_debug_plots:
#
#            index = index_outer
#
#            # run square box cut
#            lp._cut_box_square()
#
#            # plot r, t
#            _plot_r_t(self.p)
#
#            # Local Cartesian grid which the linear solution was calculated on, meshgrid version
#            X, Y = np.meshgrid(lp.x_cut,lp.y_cut)
#
#            print("Y MIN MAX", np.min(Y),np.max(Y))
#
#            # grab positive x side of linear solution
#            x_len = len(lp.x_cut) // 2
#
#            # restrict box to area considered
#            X_rest = X[:,x_len:index]
#            Y_rest = Y[:,x_len:index]
#
#            print("Y MIN MAX", np.min(Y_rest),np.max(Y_rest))
#
#            # Find chi for linear solution, only taking positive values of x
#            # The "index" value also makes it so we do not go outside the linear box
#            linear_profile = lp.cut_rho[:,x_len:index] / np.sqrt(np.abs(X_rest)) ###
#            chi0_rho = linear_profile[:,-1]  * (gamma+1) * beta_p / 2**(3/4) 
#
#            # extract chi from velocity perturbations
#            v_unit = (self.p.c_s_planet*(self.p.m_planet/self.p.m_thermal))
#            Rp     = self.p.r_planet
#            cw     = -self.p.a_cw
#            hr     = self.p.hr_planet
#            q      = self.p.q
#            gamma  = self.p.gamma
#            csp    = self.p.c_s_planet
#            p      = self.p.p
#            l      = self.p.l
#            linear_profile_v_r   = v_unit * (lp.cut_v_r[:,x_len:index]   / np.sqrt(np.abs(X_rest))) ###
#            linear_profile_v_phi = v_unit * (lp.cut_v_phi[:,x_len:index] / np.sqrt(np.abs(X_rest))) ###
#            R0  = np.sqrt((Rp + l*X_rest[:,-1])**2 + (l*Y_rest[:,-1])**2)
#            Lfu = _Lambda_fu(R0, Rp, csp, hr, gamma, q, p)
#            Lfv = _Lambda_fv(R0, Rp, csp, hr, gamma, q, p)
#            chi0_v_r = linear_profile_v_r[:,-1] / (np.sign(R0 - Rp) * Lfu)
#            chi0_v_phi = linear_profile_v_phi[:,-1] / (np.sign(R0 - Rp) * Lfv * (-cw))
#
#            # Find etas for the box using IC method
#            linear_eta = Y_rest + np.sign(X_rest) * 0.5 * X_rest**2
#
#            # Find etas using proper transformations
#            #hp = self.p.hr * self.p.r_planet
#            hp = self.p.hr_planet * self.p.r_planet
#            x_glob = 2 * hp * X_rest / 3.
#            y_glob = 2 * hp * Y_rest / 3.
#            r_glob = x_glob + self.p.r_planet
#            phi_glob = y_glob / self.p.r_planet
#            eta_lin = np.zeros(linear_profile.shape)
#            t_lin = np.zeros(linear_profile.shape)
#            for i in range(eta_lin.shape[0]):
#                #print(str(i))
#                for j in range(eta_lin.shape[1]):
#                    eta_lin[i,j] = _Eta(r_glob[i,j], phi_glob[i,j], self.p.r_planet, self.p.hr_planet, self.p.q, -self.p.a_cw)
#                    t_lin[i,j] = _t(r_glob[i,j], self.p.r_planet, self.p.hr_planet, self.p.q, self.p.p)
#
#            # Plot Chi vs eta when using eta transformation used for IC cut out
#
#            #for i in range(0, len(linear_eta[0,:]), 10):
#            #    plt.plot(linear_eta[:,i], linear_profile[:,i])
#            #    plt.plot(eta_lin[:,i], linear_profile[:,i])
#            #    plt.show()
#
#            plt.plot(eta_lin[:,-1], chi0_rho, label="chi rho")
#            plt.plot(eta_lin[:,-1], chi0_v_r, label="chi vr")
#            plt.plot(eta_lin[:,-1], chi0_v_phi, label="chi vphi")
#            plt.legend(loc="lower left")
#            plt.xlim(-10,10)
#            plt.xlabel('$\eta$')
#            plt.ylabel('$\chi$')
#            plt.show()
#
#            if True: # save results to file for plotting
#
#                name = "sim_further"
#                eta_save = eta_lin[:,-1]
#                chi_save = np.array([chi0_rho, chi0_v_r, chi0_v_phi])
#                np.save(f"{name}_eta.npy", eta_save)
#                np.save(f"{name}_chi.npy", chi_save)
#
#            plt.plot(linear_eta[:,-1], linear_profile[:,-1], label="Approximate $\eta$ transformation")
#            plt.plot(eta_lin[:,-1], linear_profile[:,-1], label="Full $\eta$ transformation")
#            plt.plot(self.eta, self.profile, ls='--', label="Actual IC used")
#            plt.legend(loc="lower left")
#            plt.xlim(-10,10)
#            plt.xlabel('$\eta$')
#            plt.ylabel('$\chi$')
#            plt.show()
#
#            plt.scatter(t_lin, eta_lin)
#            plt.show()
#
#            #eta_lin = linear_eta    ### TURN THIS ON AND OFF TO CHANGE ETA TRANSFORMATION --> COMMENT OUT TO USE INTEGRAL TRANSFORM
#
#            # get grid regular in (t,eta) to interpolate onto
#            dt_lin = np.diff(np.sort(t_lin.flatten())).mean()       # find time step interval to use on regular grid from t_lin
#                                                                    # takes minimum separation between points in original t_lin grid
#            t_lin_reg = np.arange(0, self.t0, dt_lin)
#            eta_lin_reg = copy(self.eta)
#            T_lin_reg, ETA_lin_reg = np.meshgrid(t_lin_reg, eta_lin_reg)
#
#            # interpolate solution over irregular grid onto regular grid
#            print("Interpolating Linear solution into (t, eta) space")
#            lin_solution = griddata(
#                (t_lin.flatten(), eta_lin.flatten()), 
#                linear_profile.flatten(),
#                (T_lin_reg, ETA_lin_reg),
#                method='linear'
#            )
#
#            #plt.plot(linear_eta[:,-1], linear_profile[:,-1])       # approx
#            plt.plot(eta_lin[:,-1], linear_profile[:,-1])           # integral trans
#            plt.plot(eta_lin_reg, lin_solution[:,-1])               # interpolated integral trans
#            #plt.plot(self.eta, self.profile, c='k')                # IC
#            plt.show()
#
#            # plotting (for debugging)
#            _, ax = plt.subplots()
#            myplot = ax.contourf(t_lin_reg, eta_lin_reg, np.nan_to_num(lin_solution), levels=np.arange(-4,4,0.1), cmap='RdBu', extend='both')
#            plt.colorbar(myplot)
#            plt.show()
#
#            # stick this into the solutions array, update later code so that solutions array builds on this one
#            self.linear_solution = np.nan_to_num(lin_solution)
#            self.linear_t = t_lin_reg
#
#        else:
            
        self.linear_solution = 0
        self.linear_t = 0

    # alternative IC extraction using edge of box
    def _extract_ICs_ann(self, LinearPerts: '_LinearPerts') -> None:
        """Alternate initial condition extraction where the IC is read from the edges of the box as included in the final
        solution. Usually, far more y-extent of the linear regime is used than this. Using this method will invalidate the
        solution and is meant for developer use only.
        """
        
        # grab linear perturbations object
        lp = LinearPerts

        # mass unit
        beta_p = self.p.m_planet / self.p.m_thermal

        # grab radius and phi values for edge of box
        phi_IC_outer = lp.PHI_ann[:,-1]
        phi_IC_inner = lp.PHI_ann[:,0]
        r_IC_outer   = lp.r_ann[-1]
        r_IC_inner   = lp.r_ann[ 0]

        # edge of box in local coords
        x_box_outer = (r_IC_outer - self.p.r_planet) / self.p.l
        x_box_inner = (r_IC_inner - self.p.r_planet) / self.p.l

        self.profile_outer = (lp.pert_rho_ann[:,-1] / beta_p) / np.sqrt(np.abs(x_box_outer))
        self.profile_inner = (lp.pert_rho_ann[:, 0] / beta_p) / np.sqrt(np.abs(x_box_inner))

        if False:
         plt.plot(phi_IC_outer, self.profile_outer, label="outer")
         plt.plot(phi_IC_inner, self.profile_inner, label="inner")
         plt.legend(loc="best")
         plt.show()

        # find t points
        t_IC_outer = _t(r_IC_outer, self.p.r_planet, self.p.hr_planet, self.p.q, self.p.p)
        t_IC_inner = _t(r_IC_inner, self.p.r_planet, self.p.hr_planet, self.p.q, self.p.p)

        # initialise arrays for corresponding eta points
        self.eta_outer = np.zeros(len(phi_IC_outer))
        self.eta_inner = np.zeros(len(phi_IC_outer))

        # perform transformation
        for i in range(len(phi_IC_outer)):
            self.eta_outer[i] = _Eta(r_IC_outer, phi_IC_outer[i], self.p.r_planet, self.p.hr_planet, self.p.q, -1)
            self.eta_inner[i] = _Eta(r_IC_inner, phi_IC_inner[i], self.p.r_planet, self.p.hr_planet, self.p.q, -1)

        # set t0
        self.t0_outer = t_IC_outer
        self.t0_inner = t_IC_inner

        # set eta_tilde for outer wake:
        for i in range(len(self.eta_outer)):
            if self.profile_outer[i] == 0 and self.eta_outer[i] > -10 and self.eta_outer[i] < 0:
                zero_outer = self.eta_outer[i]
            elif i!= (len(self.eta_outer) - 1) and self.profile_outer[i] * self.profile_outer[i + 1] < 0 and self.eta_outer[i] > -10 and self.eta_outer[i] < 0:
                zero_outer = 0.5 * (self.eta_outer[i] + self.eta_outer[i + 1])
        self.eta_tilde_outer = -zero_outer

        # set eta_tilde for inner wake:
        for i in range(len(self.eta_inner)):
            if self.profile_inner[i] == 0 and self.eta_inner[i] > -10 and self.eta_inner[i] < 0:
                zero_inner = self.eta_inner[i]
            elif i!= (len(self.eta_inner) - 1) and self.profile_inner[i] * self.profile_inner[i + 1] < 0 and self.eta_inner[i] > -10 and self.eta_inner[i] < 0:
                zero_inner = 0.5 * (self.eta_inner[i] + self.eta_inner[i + 1])
        self.eta_tilde_inner = -zero_inner

        # set C for outer wake:
        deta_outer = self.eta_outer[1] - self.eta_outer[0]
        profile0_outer = self.profile_outer[self.eta_outer < -self.eta_tilde_outer]
        C0_outer = -np.trapz(profile0_outer, dx = deta_outer)
        self.C_outer = (self.p.gamma + 1) * (self.p.m_planet / self.p.m_thermal) * C0_outer / 2**(3/4)

        # set C for inner wake:
        deta_inner = self.eta_inner[1] - self.eta_inner[0]
        profile0_inner = self.profile_inner[self.eta_inner < -self.eta_tilde_inner]
        C0_inner = -np.trapz(profile0_inner, dx = deta_inner)
        self.C_inner = (self.p.gamma + 1) * (self.p.m_planet / self.p.m_thermal) * C0_inner / 2**(3/4)

        #print('     Outer Wake:')
        #print('         eta_tilde = ', self.eta_tilde_outer)
        #print('         C0 = ', C0_outer)
        #print('         t0 = ', self.t0_outer)
        #print('     Inner Wake:')
        #print('         eta_tilde = ', self.eta_tilde_inner)
        #print('         C0 = ', C0_inner)
        #print('         t0 = ', self.t0_inner)

        # hard to explain, but they're not needed so set to zero
        self.linear_solution = 0
        self.linear_t = 0

    # propagate the wake using IC
    def _get_non_linear_perts(self) -> None:
        """Calculate the perturbations in the non-linear regime by propagating the wake away from the planet
        by solving Burger's eqn and then transforming the results to physical coordinates.
        """

        beta_p = self.p.m_planet / self.p.m_thermal

        print('Propagating outer wake... ')

        timer_0 = time.perf_counter()

        time_outer, eta_outer, solution_outer = _solve_burgers(
            self.eta_outer, 
            self.profile_outer, 
            self.p.gamma, 
            beta_p, 
            self.C_outer, 
            self.p.CFL, 
            self.eta_tilde_outer, 
            self.t0_outer, 
            self.linear_solution, 
            self.linear_t, 
            self.p.show_teta_debug_plots,
            self.p.tf_fac
        )

        timer_1 = time.perf_counter()

        print(f'Completed in {timer_1-timer_0:0.2f} s')

        print('Propagating inner wake... ')

        timer_0 = time.perf_counter()

        time_inner, eta_inner, solution_inner = _solve_burgers(
            self.eta_inner, 
            -self.profile_inner,     # this minus sign is intended
            self.p.gamma, 
            beta_p, 
            self.C_inner, 
            self.p.CFL, 
            self.eta_tilde_inner, 
            self.t0_inner, 
            self.linear_solution, 
            self.linear_t, 
            self.p.show_teta_debug_plots,
            self.p.tf_fac
        )

        timer_1 = time.perf_counter()

        print(f'Completed in {timer_1-timer_0:0.2f} s')

        # final time of solutions before N-wave behaviour
        tf_outer = time_outer[-1]
        tf_inner = time_inner[-1]

        # needed constants from solutions, inner and outer
        eta_tilde_outer = self.eta_tilde_outer
        eta_tilde_inner = self.eta_tilde_inner
        C_outer         = self.C_outer
        C_inner         = self.C_inner
        t0_outer        = self.t0_outer
        t0_inner        = self.t0_inner

        # parameters of run
        Rp      = self.p.r_planet
        x_match = 2*self.p.scale_box 
        l       = self.p.l
        cw      = -self.p.a_cw
        hr      = self.p.hr_planet
        q       = self.p.q
        csp     = self.p.c_s_planet
        p       = self.p.p

        # physical parameters
        gamma = self.p.gamma


        # if using a Cartesian grid

        timer_0 = time.perf_counter()

        if self.g.info["Type"] == "cartesian":
            print("Mapping to Physical Coords")

            x = self.g.x 
            y = self.g.y

            [x_grid, y_grid] = np.meshgrid(x, y)
            r_grid = np.sqrt(x_grid**2 + y_grid**2)
            pphi_grid = np.arctan2(y_grid, x_grid)

            tt = _t_vector(r_grid, Rp, hr, q, p)

            CChi = _get_chi_vector(
                pphi_grid, 
                r_grid,
                tt,
                time_outer,
                time_inner, 
                eta_outer, 
                eta_inner, 
                eta_tilde_outer,
                eta_tilde_inner, 
                C_outer,
                C_inner, 
                solution_outer, 
                solution_inner, 
                t0_outer,
                t0_inner, 
                tf_outer,
                tf_inner, 
                Rp, 
                x_match, 
                l, 
                cw, 
                hr, 
                q, 
                p,
            )

            # COMPUTE DENSITY AND VELOCITY PERTURBATIONS
            dnl, unl, vnl = _get_dens_vel(
                np.transpose(r_grid), 
                np.transpose(CChi), 
                gamma, 
                Rp, 
                cw, 
                csp, 
                hr, 
                q, 
                p
            )

        # if using a cylindrical grid
        else:
            r = self.g.r
            phi = self.g.phi

            r_grid, pphi_grid = np.meshgrid(r, phi)

            tt = _t_vector(r_grid, Rp, hr, q, p)

            Chi = _get_chi_vector(
                pphi_grid, 
                r_grid, 
                tt,
                time_outer,
                time_inner, 
                eta_outer, 
                eta_inner, 
                eta_tilde_outer,
                eta_tilde_inner, 
                C_outer,
                C_inner, 
                solution_outer, 
                solution_inner, 
                t0_outer,
                t0_inner, 
                tf_outer,
                tf_inner, 
                Rp, 
                x_match, 
                l, 
                cw, 
                hr, 
                q, 
                p
            )

            # COMPUTE DENSITY AND VELOCITY PERTURBATIONS
            dnl, unl, vnl = _get_dens_vel(
                r_grid, 
                Chi, 
                gamma, 
                Rp, 
                cw, 
                csp, 
                hr, 
                q, 
                p
            )
        
        timer_1 = time.perf_counter()
        print(f'Completed in {timer_1-timer_0:0.2f} s')

        self.rho = dnl
        self.vr = 1e-5*unl
        self.vphi = 1e-5*vnl


# __init__.py
# Written by Thomas Hilder

"""
Wakeflow allows users to generate and manipulate semi-analytic models of planet wakes.
"""

# wakeflow package versionss
__version__ = "1.2.0"

# give users access to the WakeflowModel class
from .wakeflow import WakeflowModel

# give users access to the Grid class, realistically they shouldn't need it but it 
# may be desirable by advanced users or the developers
from .grid import _Grid


# linear_perts.py
# Written by Thomas Hilder and Francesco Bollati

"""
Contains the LinearPerts class responsible for handling the linear regime of the models.
"""

# type hinting without circular imports
from typing                 import TYPE_CHECKING

import sys, pkg_resources, tarfile
import numpy                    as np
import matplotlib.pyplot        as plt
from scipy.interpolate      import RectBivariateSpline

if TYPE_CHECKING:
    from .model_setup       import _Parameters

# NOTE: contents are intended for internal use and should not be directly accessed by users

# class for storing the results from the linear regime
class _LinearPerts():
    """
    Class to store the results from the linear regime nearby the planet.
    """

    # intitalise object given parameters. locate linear results data and read in
    def __init__(self, parameters : '_Parameters') -> None:

        # grab parameters object
        self.p = parameters

        # get location of linear perturbations data files
        pert_loc = pkg_resources.resource_filename('wakeflow', 'data/linear_perturbations.npy')
        mesh_loc = pkg_resources.resource_filename('wakeflow', 'data/linear_perturbations_mesh.npy')

        # read perturbations from files
        try:
            perts = np.load(pert_loc)
            mesh  = np.load(mesh_loc)

        # in the case files have not been extracted from tarballs yet, extract them
        except FileNotFoundError:

            # open tarballs
            pert_tar = tarfile.open(f"{pert_loc}.tar.gz")
            mesh_tar = tarfile.open(f"{mesh_loc}.tar.gz")

            # extract npy files
            loc = pkg_resources.resource_filename('wakeflow', 'data')
            pert_tar.extractall(loc)
            mesh_tar.extractall(loc)

            # close tarballs
            pert_tar.close()
            mesh_tar.close()

            # read perturbations from files
            perts = np.load(pert_loc)
            mesh  = np.load(mesh_loc)

        # get perturbation arrays
        self.pert_v_r   = perts[0]
        self.pert_v_phi = perts[1]
        self.pert_rho   = perts[2]

        # grid
        self.X = mesh[0]
        self.Y = mesh[1]

        # linear perturbations read in grid
        x = self.X[0,:]
        y = self.Y[:,0]

        # define constants for linear perts
        self.x_box = 2 * self.p.scale_box
        
        # cut square box grid in linear regime
        self.x_cut = x[np.argmin(x < -self.x_box) : np.argmin(x < self.x_box) + 1]
        self.y_cut = y[np.argmin(y < -self.x_box) : np.argmin(y < self.x_box) + 1]

        # test plot
        if False:
            plt.figure()
            plt.contourf(x, y, self.pert_rho, levels=100, cmap='RdBu')
            plt.xlim(self.x_cut[0],self.x_cut[-1])
            plt.ylim(self.y_cut[0],self.y_cut[-1])
            plt.colorbar()
            plt.show()

    # old method of extracting linear perturbations 
    def _cut_box_square(self) -> None:
        """Deprecated. This is only used in the case where you want to plot the linear solution in t, eta space
        """

        # box size (in units of Hill radius), with default scale_box = 1. (note for conversions that self.p.l = 1 Hill radius in cgs)
        box_size = 2*self.p.scale_box
        artificial_y_scale = 6

        # linear perturbations read in grid
        x = self.X[0,:]
        y = self.Y[:,0]

        # cut square box grid in linear regime
        x_cut = x[np.argmin(x < -box_size)                    : np.argmin(x < box_size)                    + 1]
        y_cut = y[np.argmin(y < -artificial_y_scale*box_size) : np.argmin(y < artificial_y_scale*box_size) + 1]

        # find cut indicies 
        x_cut_i1 = np.argmin(x < -box_size)
        x_cut_i2 = np.argmin(x <  box_size) + 1
        y_cut_i1 = np.argmin(y < -artificial_y_scale * box_size)
        y_cut_i2 = np.argmin(y <  artificial_y_scale * box_size) + 1

        # cut perturbation arrays
        cut_v_r   = self.pert_v_r  [y_cut_i1:y_cut_i2, x_cut_i1:x_cut_i2]
        cut_v_phi = self.pert_v_phi[y_cut_i1:y_cut_i2, x_cut_i1:x_cut_i2]
        cut_rho   = self.pert_rho  [y_cut_i1:y_cut_i2, x_cut_i1:x_cut_i2]

        self.cut_rho   = cut_rho
        self.cut_v_r   = cut_v_r
        self.cut_v_phi = cut_v_phi

        # scale to cgs units (and account for rotation direction)
        self.pert_v_r_sq   = cut_v_r   * self.p.c_s_planet * (self.p.m_planet/self.p.m_thermal)
        self.pert_v_phi_sq = cut_v_phi * self.p.c_s_planet * (self.p.m_planet/self.p.m_thermal)
        self.pert_rho_sq   = cut_rho   *                     (self.p.m_planet/self.p.m_thermal)

        self.pert_rho_sq_unscaled = cut_rho # this does not need to account for direction

        # account for rotation direction
        if self.p.a_cw == -1:
            self.pert_v_r_sq   =  np.flipud(self.pert_v_r_sq)
            self.pert_v_phi_sq = -np.flipud(self.pert_v_phi_sq)
            self.pert_rho_sq   =  np.flipud(self.pert_rho_sq)

        # save grids
        self.x_cut = x_cut 
        self.y_cut = y_cut
        self.x_sq  = self.p.l * x_cut + self.p.r_planet
        self.y_sq  = self.p.l * y_cut
        
    # extract linear perturbations and interpolate onto annulus segment grid for global results
    def _cut_box_annulus_segment(self) -> None:
        """Extract the part of the linear solution needed for the model, and interpolate onto appropriate grid
        """

        # box size (in units of Hill radius), note for conversions that self.p.l = 1 Hill radius in cgs
        x_box_size = 2*self.p.scale_box
        y_box_size = 2*self.p.scale_box_ang

        # linear perturbations read in grid
        x = self.X[0,:]
        y = self.Y[:,0]

        # cut square box in linear regime
        x_cut = x[np.argmin(x < -x_box_size) : np.argmin(x < x_box_size) + 1]
        y_cut = y[np.argmin(y < -y_box_size) : np.argmin(y < y_box_size) + 1]

        # annulus segment grid, granularity from square box
        r = np.linspace(
            self.p.r_planet - x_box_size*self.p.l, 
            self.p.r_planet + x_box_size*self.p.l, 
            len(x_cut)
        )
        y_ = np.linspace(
            -y_box_size*self.p.l,
            y_box_size*self.p.l,
            len(y_cut)
        )

        R, Y_ = np.meshgrid(r, y_)

        # for the points on our annulus segment, use transformations to find the x,y values on the original perturbations grid
        # if the user has chosen a linear box with very large angular extent, R and Y will be equal at some point, creating an error
        with np.errstate(all='raise'):
            try:

                # pick x values for the box
                if self.p.box_warp:
                    X_pert_grid = (R - self.p.r_planet) / self.p.l
                else:
                    X_pert_grid = (np.sqrt(R**2 - Y_**2) - self.p.r_planet) / self.p.l

                # get phi values
                PHI = np.arctan2(Y_, np.sqrt(R**2 - Y_**2))

            except FloatingPointError:
                print("Error: Reduce 'scale_box_ang' parameter.")
                print("Exiting")
                sys.exit(1)
        
        # unscale Y
        Y_pert_grid = Y_ / self.p.l

        # cut big perturbations grid to just outside annulus
        self.r_min = self.p.r_planet - x_box_size*self.p.l
        self.r_max = self.p.r_planet + x_box_size*self.p.l

        x_min_global = np.sqrt(self.r_min**2 - (y_box_size*self.p.l)**2)
        x_min_local  = (x_min_global - self.p.r_planet) / self.p.l

        # find cut indicies (remember we need to scale back to units of Hill radius )
        x_cut_i1 = np.argmin(x <  x_min_local)
        x_cut_i2 = np.argmin(x <  x_box_size) + 1
        y_cut_i1 = np.argmin(y < -y_box_size)
        y_cut_i2 = np.argmin(y <  y_box_size) + 1

        # cut grid
        x_int_cut = x[x_cut_i1 : x_cut_i2]
        y_int_cut = y[y_cut_i1 : y_cut_i2]

        # cut perturbation arrays
        cut_v_r   = self.pert_v_r   [y_cut_i1:y_cut_i2, x_cut_i1:x_cut_i2]
        cut_v_phi = self.pert_v_phi [y_cut_i1:y_cut_i2, x_cut_i1:x_cut_i2]
        cut_rho   = self.pert_rho   [y_cut_i1:y_cut_i2, x_cut_i1:x_cut_i2]

        if False:
            plt.contourf(x_int_cut, y_int_cut, cut_rho, cmap="RdBu", vmin=-1, vmax=1, levels=100)
            plt.show()

        # account for rotation direction
        if self.p.a_cw == -1:
            cut_v_r   =  np.flipud(cut_v_r)
            cut_v_phi = -np.flipud(cut_v_phi)
            cut_rho   =  np.flipud(cut_rho)

        # interpolation over cut (cartesian) grid
        interp_v_r   = RectBivariateSpline(y_int_cut, x_int_cut, cut_v_r)
        interp_v_phi = RectBivariateSpline(y_int_cut, x_int_cut, cut_v_phi)
        interp_v_rho = RectBivariateSpline(y_int_cut, x_int_cut, cut_rho)

        # evaluate interpolation over our annulus segment
        self.pert_v_r_ann   = interp_v_r.ev  (Y_pert_grid, X_pert_grid)
        self.pert_v_phi_ann = interp_v_phi.ev(Y_pert_grid, X_pert_grid)
        self.pert_rho_ann   = interp_v_rho.ev(Y_pert_grid, X_pert_grid)

        if False:
            plt.imshow(self.pert_rho_ann, cmap="RdBu", vmin=-1, vmax=1)
            plt.show()

        # scale to cgs units
        self.pert_v_r_ann   *= self.p.c_s_planet*(self.p.m_planet/self.p.m_thermal)
        self.pert_v_phi_ann *= self.p.c_s_planet*(self.p.m_planet/self.p.m_thermal)
        self.pert_rho_ann   *=                   (self.p.m_planet/self.p.m_thermal)

        # scale velocities to km/s
        self.pert_v_r_ann   *= 1e-5
        self.pert_v_phi_ann *= 1e-5

        # save annulus grid
        self.r_ann   = r
        self.y_ann   = y_
        self.R_ann   = R
        self.PHI_ann = PHI

        if False:
            # plotting (for debugging)
            _, ax = plt.subplots(subplot_kw=dict(projection='polar'))
            myplot = ax.contourf(PHI, R, self.pert_rho_ann/(self.p.m_planet/self.p.m_thermal), levels=300, vmin=-1, vmax=1, cmap='RdBu')
            ax.set_ylim(0, self.p.r_outer)
            plt.colorbar(myplot)
            plt.show()



# mcfost_interface.py
# Written by Thomas Hilder

"""
Contains functions for generating MCFOST grid information and writing MCFOST parameter files.
Requires pymcfost which may be found here: https://github.com/cpinte/pymcfost
"""

import os, subprocess, pkg_resources
import numpy          as np
from .model_setup import _Parameters
from typing       import Tuple

try:
    from pymcfost.parameters        import Params
    from pymcfost.disc_structure    import Disc
except ImportError:
    pass

# NOTE: contents are intended for internal use and should not be directly accessed by users

# call pymcfost to write a parameter file for MCFOST
def _make_mcfost_parameter_file(parameters: _Parameters) -> None:
    """Take Wakeflow parameters and write an MCFOST .para file by calling pymcfost.
    """
    
    # grab run parameters
    p = parameters

    # get location of mcfost reference para file
    mcfost_para_file = pkg_resources.resource_filename('wakeflow', 'data/ref3.0_3D.para')

    # setting up an mcfost parameters object
    mp = Params(mcfost_para_file)

    # --- mcfost parameters that are not set by config.yaml

    mp.phot.nphot_T     = 1.28e+07      # set number of photons for temp. calculation
    mp.simu.compute_SED = False         # don't compute SED

    mp.map.nx   = 1001                  # output resolution
    mp.map.ny   = 1001
    mp.map.size = 1200

    mp.simu.image_symmetry   = False    # we don't have any symmetries
    mp.simu.central_symmetry = False
    mp.simu.axial_symmetry   = False

    mp.mol.molecule[0].n_trans = 1      
    mp.grid.n_rad_in           = 1            

    mp.mol.v_turb = 0.0                 # no turbulence

    mp.zones[0].edge = 0.0              # not using smoothed inner and outer edge

    mp.map.RT_n_az   = 1                # we are only using one azimuth
    mp.map.RT_ntheta = 1                # we are only using one inclination

    # --- mcfost parameters that are set by config.yaml

    mp.grid.n_rad    = p.n_r            # grid geometry
    mp.grid.n_az     = p.n_phi
    mp.grid.nz       = p.n_z
    mp.zones[0].Rin  = p.r_inner
    mp.zones[0].Rout = p.r_outer

    mp.zones[0].h0          = p.h_ref   # scale height at reference radius
    mp.zones[0].Rref        = p.r_ref   # reference radius
    mp.zones[0].Rc          = p.r_c     # critical radius unused so set to 0
    mp.zones[0].flaring_exp = p.beta    # flaring index: h \propto r^beta

    mp.map.distance = p.distance        # distance

    mp.stars[0].Teff = p.temp           # star temperature

    mp.map.PA = p.PA                    # PA angle

    mp.map.RT_imin = p.inclination      # inclination
    mp.map.RT_imax = p.inclination

    mp.map.RT_az_min = p.PAp            # azimuth used to rotate to correct planet position
    mp.map.RT_az_max = p.PAp

    mp.mol.molecule[0].v_max = p.v_max  # max velocity
    mp.mol.molecule[0].nv    = p.n_v    # number of velocity channels

    # write mcfost parameter file
    mp.writeto(f"{p.system}/{p.name}/mcfost/mcfost_{p.name}.para")

# Call MCFOST to generate the grid data which will be read into Wakeflow
def _make_mcfost_grid_data(parameters: _Parameters) -> None:
    """Take grid parameters specified by Wakeflow user and call MCFOST to generate grid data.
    """

    print("Generating MCFOST grid data...")

    # grab run parameters
    p = parameters

    # generate grid data by running mcfost (in mcfost directory)
    working_dir = os.getcwd()
    os.chdir(f"{p.system}/{p.name}/mcfost/")
    subprocess.call(["rm", "-rf", "data_disk", "data_disk_old"])
    subprocess.call(["mcfost", f"mcfost_{p.name}.para", "-disk_struct"]) #, stdout=subprocess.DEVNULL)
    os.chdir(working_dir)

    print("Done")

# read in MCFOST grid data previously generated 
def _read_mcfost_grid_data(parameters: _Parameters) -> Tuple[np.ndarray, np.ndarray]:
    """Read in grid data generated by MCFOST and return radial and z coordinates.
    """

    # grab run parameters
    p = parameters
    
    # reading disk data
    #with HiddenPrints():
    mcfost_disk = Disc(f"./{p.system}/{p.name}/mcfost/")

    # getting radii and heights
    r = mcfost_disk.r()[:, p.n_z:, :]
    z = mcfost_disk.z()[:, p.n_z:, :]

    return r, z

#class HiddenPrints:
#    def __enter__(self):
#        self._original_stdout = sys.stdout
#        sys.stdout = open(os.devnull, 'w')
#
#    def __exit__(self, exc_type, exc_val, exc_tb):
#        sys.stdout.close()
#        sys.stdout = self._original_stdout


# wakeflow.py
# Written by Thomas Hilder

"""
Contains the WakeflowModel class, intended for use by users to generate, configure and run models of planet wakes.
"""

import subprocess, os
from .model_setup         import _load_config_file, _run_setup, _Parameters
from .grid                import _Grid
from .linear_perts        import _LinearPerts
from .non_linear_perts    import _NonLinearPerts
#from phantom_interface   import PhantomDump

# class for use by the user to interact with Wakeflow
class WakeflowModel():
    """
    Model object allowing you to configure, generate and save planet wake model results.

    Attributes
    ----------
    model_params : dict
        Dictionary containing the parameters to be used in the model as specified by the user.
    default_params : dict
        Dictionary containing the default WakeflowModel parameters.
    """

    # instantiate class
    def __init__(self) -> None:
        """ Instantiate model
        """

        print("Model initialised.")

    # set the parameters for the model
    def configure(
        self,
        name:                 str = "default_results_dir",
        system:               str = "default_parent_dir",
        m_star:             float = 1.0,
        m_planet:           float = 0.1,
        r_outer:            float = 500,
        r_inner:            float = 100,
        r_planet:           float = 250,
        r_ref:              float = None,
        r_c:                float = 0,
        q:                  float = 0.25,
        p:                  float = 1.0,
        hr:                 float = 0.10,
        dens_ref:           float = 1.0,
        cw_rotation:         bool = False,
        grid_type:          float = "cartesian",
        n_x:                  int = 400,
        n_y:                  int = 400,
        n_r:                  int = 200,
        n_phi:                int = 160,
        n_z:                  int = 50,
        r_log:               bool = False,
        make_midplane_plots: bool = True,
        show_midplane_plots: bool = True,
        dimensionless:       bool = False,
        include_planet:      bool = True,
        include_linear:      bool = True,
        save_perturbations:  bool = True,
        save_total:          bool = True,
        write_FITS:          bool = False,
        run_mcfost:          bool = False,
        inclination:        float = -225,
        PA:                 float = 45,
        PAp:                float = 45,
        temp_star:          float = 9250,
        distance:           float = 101.5,
        v_max:              float = 3.2,
        n_v:                float = 40
    ) -> None:
        """
        Configure the model according the specifications by the user.

        Parameters
        ----------
        name : str 
            results are saved a directory called name.
        system : str
            name of parent directory of results.
        m_star : float
            central star mass in solar masses.
        m_planet : float
            planet mass in Jupiter masses. Can also give list(float) to run multiple planet masses.
        r_outer : float
            outer disk radius in au.
        r_inner : float
            inner disk radius in au.
        r_planet : float
            orbital radius of planet in au.
        r_ref : float
            reference radius r_ref in au.
        r_c : float
            critical radius r_c in au, used for exponentially tapered density profile. ignored if set to 0.
        q : float 
            q index for sound speed profile, defined as c_s \propto r^{-q}.
        p : float
            p index for density profile (NOT surface density), defined as rho \propto r^{-p} or \propto r^{-p}*\exp{-r/r_c}^{2-p}.
        hr : float 
            disk aspect ratio at r_ref.
        dens_ref : float
            density at r_ref in g/cm^3.
        cw_rotation : bool
            clockwise disk rotation (True) or anticlockwise disk rotation (False).
        grid_type : float
            grid type, either "cylindrical" or "cartesian" or "mcfost". Must choose "mcfost" to write a .FITS file.
        n_x : int
            number of grid points in x.
        n_y : int
            number of grid points in y.
        n_r : int
            number of grid points in radius.
        n_phi : int
            number of grid points in azimuth.
        n_z : int
            number of grid points in z.
        r_log : bool
            True for logarithmically spaced radii, False for linear. Ignored for "mcfost" grid type.
        make_midplane_plots : bool
            Create midplane plots of density and velocity perturbations?
        show_midplane_plots : bool
            Display midplane plots of density and velocity perturbations to user?
        dimensionless : bool
            True for dimensionless results otherwise False.
        include_planet : bool
            Include the planet-induced perturbations? Set to False to generate unperturbed disk.
        include_linear : bool
            Include the results from the linear regime?
        save_perturbations : bool
            Save the perturbations?
        save_total : bool
            Save the totals (perturbations + background disk)?
        write_FITS : bool
            Generate a .FITS file to run in MCFOST? Requires "mcfost" grid type.
        run_mcfost : bool
            Call MCFOST to generate line emission cubes on results, requires write_FITS=True.
        inclination : float
            Inclination in degrees for MCFOST.
        PA : float
            Position angle in degrees for MCFOST.
        PAp : float
            Planet position in degrees angle for MCFOST.
        temp_star : float
            Star temperature in Kelvin for MCFOST.
        distance : float
            System distance in parsecs for MCFOST.
        v_max : float
            maxmimum magnitude velocity for cube in MCFOST.
        n_v : float
            number of velocity channels for MCFOST.
        """

        # developer parameters, intentionally not easy to change as can very easily invalidate results
        # 
        adiabatic_index       = 1.6666667     # adiabatic index
        damping_malpha        = 0.0           # artificial damping NOT IMPLEMENTED
        CFL                   = 0.5           # Courant stability factor (require <0.5)
        scale_box             = 1.0           # linear box length scale factor in radial direction
        scale_box_ang         = 1.0           # linear box length scale factor in angular direction
        tf_fac                = 1.0           # scale factor for t coordinate where self-similar solution is used
        show_teta_debug_plots = False         # show (t,eta,chi) space developer plots
        box_warp              = True          # interpret y coordinate of linear regime as arc length, or truly vertical? True (default) for former
        use_box_IC            = False         # use only part of linear regime in box as initial condition for non-linear evolution

        # generate dictionary for model parameters by grabbing all local variables
        self.model_params = locals()

        # remove "self" item
        del self.model_params["self"]

        # confirmation message
        print("Model configured.")

    # set the parameters for the model by handing a .yaml file
    def configure_from_file(
        self,
        param_file: str
    ) -> None:
        """
        Configure the model by reading in .yaml file provided by user.

        Parameters
        ----------
        param_file : str 
            path to .yaml file provided by user specifying run parameters.
        """

        # generate default parameters
        self.configure()
        self.default_params = self.model_params
        del self.model_params

        # read in file to dictionary
        self.model_params = _load_config_file(param_file, self.default_params)

        # confirmation message
        print(f"Model configuration read from file: {param_file}")

    # generate the model using the configuration specified by the user
    def run(self, overwrite: bool = False) -> None:
        """
        Generate results for model, requires user to have called either configure or configure_from_file first.

        Parameters
        ----------
        overwrite : bool 
            Overwrite previous results with identical name?
        """
        
        print(
            "                 __        ______             \n _      ______ _/ /_____  / __/ /___ _      __\n| | /| / / __ `/ //_/ _ \/ /_/ / __ \ | /| / /\n| |/ |/ / /_/ / ,< /  __/ __/ / /_/ / |/ |/ / \n|__/|__/\__,_/_/|_|\___/_/ /_/\____/|__/|__/  "
        )
        
        # run setup
        try:
            params = _run_setup(self.model_params, overwrite=overwrite)
        except ArithmeticError:
            raise Exception("Model has not been configured.")

        # grab list of planet masses
        if params.m_planet is not None:
            planet_masses = [params.m_planet]
        else:
            planet_masses = params.m_planet_array

        # run wakeflow for each planet mass
        for mass_p in planet_masses:
            params.m_planet = mass_p
            print(f"\n* Creating {mass_p} Mj model:")
            self._run_wakeflow(params)

        # run mcfost for each model
        if params.run_mcfost == True:
            print("\n* Running MCFOST on each model:")
            for mass_p in planet_masses:
                print(f"{mass_p} Mj...")
                working_dir = os.getcwd()
                os.chdir(f"{params.system}/{params.name}/{mass_p}Mj/")
                subprocess.call(
                    ["mcfost", "mcfost.para", "-df", "wakeflow_model.fits", "-mol", "-freeze-out", "20", "-photodissociation", "-photodesorption"], 
                    stdout=subprocess.DEVNULL
                )
                os.chdir(working_dir)

        print("\n* Done!")

    # internal method that is called by self.run to generate the results for a specific set of parameters
    # may be called more than once if the user has specified multiple planet masses
    def _run_wakeflow(self, params: _Parameters) -> None:
        """
        Internal use method for generating the planet wake by calling other parts of Wakeflow.

        Parameters
        ----------
        params : Parameters 
            Paramaters object specfiying options for the model to be generated.
        """

        print("Generating unperturbed background disk")

        # make empty grid for unperturbed disk
        grid_background = _Grid(params)
        grid_background._make_grid()

        # fill grid with Keplerian, power law disk
        grid_background._make_keplerian_disk()

        if params.use_planet:

            print("Extracting linear perturbations nearby planet")

            # make empty grid for linear perturbations
            grid_lin_perts = _Grid(params)
            grid_lin_perts._make_grid()
            grid_lin_perts._make_empty_disk()

            # extract linear perturbations from file
            lin_perts = _LinearPerts(params)
            lin_perts._cut_box_annulus_segment()

            # add the linear perturbations onto grid
            grid_lin_perts._add_linear_perturbations(lin_perts, grid_background.rho)

            # make empty grid for non-linear perturbations
            grid_nonlin_perts = _Grid(params)
            grid_nonlin_perts._make_grid()
            grid_nonlin_perts._make_empty_disk()

            # initialise non-linear perturbations
            nonlin_perts = _NonLinearPerts(params, grid_nonlin_perts)

            # extract initial condition from the linear perturbations
            nonlin_perts._extract_ICs(lin_perts)
            if params.use_box_IC:
                nonlin_perts._extract_ICs_ann(lin_perts)

            # solve for non-linear perturbations
            nonlin_perts._get_non_linear_perts()

            # add non-linear perturbations to grid
            grid_nonlin_perts._add_non_linear_perturbations(nonlin_perts, grid_background.rho)

            # merge grids for result
            if params.include_linear:
                grid_background._merge_grids(grid_lin_perts)

            # merge grids for results
            grid_background._merge_grids(grid_nonlin_perts)

            # flip results if desired
            if params.user_cw_rotation:
                grid_background._flip_results()

            # merge grids to save or plot perturbations
            if params.make_midplane_plots or params.save_perturbations:

                if params.include_linear:
                    grid_nonlin_perts._merge_grids(grid_lin_perts)

                # flip results if desired
                if params.user_cw_rotation:
                    grid_nonlin_perts._flip_results()

                if params.dimensionless:
                    grid_nonlin_perts._remove_dimensions()

                if params.make_midplane_plots:
                    if params.show_midplane_plots:
                        print('\n* Displaying results:')
                    grid_nonlin_perts._show_disk2D(0, show=params.show_midplane_plots, save=True, dimless=params.dimensionless)

        if params.save_perturbations or params.save_total:
            print("\n* Saving results:")

        # save perturbations
        if params.save_perturbations:
            #print("Saving perturbations to file")
            grid_nonlin_perts._save_results("delta", "Perturbations")

        if params.dimensionless:
            grid_background._remove_dimensions(scale_dens=True)

        # save perts + background
        if params.save_total:
            #print("Saving background + perturbations to file")
            grid_background._save_results("total", "Total        ")

        # write fits file
        if params.write_FITS:
            grid_background._write_fits_file()


# model_setup.py
# Written by Thomas Hilder

"""
Contains the Parameters class responsible for storing the parameters used in a Wakeflow model. Additionally, contains
functions used read in said parameters, check their validity, and create/replace the results directory.
"""

import yaml, os
import shutil   as sh
import numpy    as np

# NOTE: contents are intended for internal use and should not be directly accessed by users

# class for storing physical constants, will be inherited by _Parameters class
class _Constants:
    """
    Constants class for storing physical constants needed by Wakeflow. Inherited by _Parameters.
    """

    def __init__(self):

        from astropy import constants as c

        # all constants given in CGS units
        self.m_solar    = c.M_sun.cgs.value
        self.m_jupiter  = c.M_jup.cgs.value
        self.G_const    = c.G.cgs.value
        self.au         = c.au.cgs.value

# class for storing parameters to be used in a Wakeflow model
class _Parameters(_Constants):
    """
    Class for storing Wakeflow model parameters to be easily handed around by various parts of Wakeflow.
    """

    # read in parameters from dictionary to class attributes, calculate some needed quantities
    def __init__(self, config: dict) -> None:
        """Inherit constants, read in parameters to dictionary, calculate quantities like thermal mass and scale height.
        """

        # inherit constants
        super().__init__()

        # run_info parameters
        self.name   = str(config["name"])
        self.system = str(config["system"])

        # disk parameters
        self.m_star = float(config["m_star"])

        # user can provide either single planet mass or list of planet masses
        try:
            self.m_planet       = float(config["m_planet"])
            self.m_planet_array = None
        except:
            self.m_planet       = None
            self.m_planet_array = list(config["m_planet"])

        self.r_outer  = float(config["r_outer"])
        self.r_inner  = float(config["r_inner"])
        self.r_planet = float(config["r_planet"])
        try:
            self.r_ref = float(config["r_ref"])
        except:
            self.r_ref = self.r_planet
        self.r_c      = float(config["r_c"])
        self.q        = float(config["q"])
        self.dens_p   = float(config["p"])
        self.p        = self.dens_p + self.q - 1.5
        self.hr       = float(config["hr"])
        self.rho_ref  = float(config["dens_ref"])

        # override given rotation, only calculated anticlockwise
        self.cw_rotation      = False
        self.user_cw_rotation = bool(config["cw_rotation"])

        # angles parameters
        self.inclination = float(config["inclination"])
        self.PA          = float(config["PA"])
        self.PAp         = float(config["PAp"])

        # grid parameters
        self.grid_type = str (config["grid_type"])
        self.n_x       = int (config["n_x"])
        self.n_y       = int (config["n_y"])
        self.n_r       = int (config["n_r"])
        self.n_phi     = int (config["n_phi"])
        self.n_z       = int (config["n_z"])
        self.r_log     = bool(config["r_log"])

        # plot parameters
        self.make_midplane_plots   = bool(config["make_midplane_plots"])
        self.show_midplane_plots   = bool(config["show_midplane_plots"])
        self.show_teta_debug_plots = bool(config["show_teta_debug_plots"])

        # results parameters
        self.use_planet         = bool(config["include_planet"])
        self.include_linear     = bool(config["include_linear"])
        self.save_perturbations = bool(config["save_perturbations"])
        self.save_total         = bool(config["save_total"])
        self.write_FITS         = bool(config["write_FITS"])
        self.dimensionless      = bool(config["dimensionless"])

        # mcfost parameters
        self.run_mcfost = bool (config["run_mcfost"])
        self.temp       = float(config["temp_star"])
        self.distance   = float(config["distance"])
        self.v_max      = float(config["v_max"])
        self.n_v        = int  (config["n_v"])

        # physical parameters
        self.gamma  = float(config["adiabatic_index"])
        self.malpha = float(config["damping_malpha"])

        # numerical parameters
        self.CFL           = float(config["CFL"])
        self.scale_box     = float(config["scale_box"])
        self.scale_box_ang = float(config["scale_box_ang"])
        self.box_warp      = bool (config["box_warp"])
        self.use_box_IC    = bool (config["use_box_IC"])
        self.tf_fac        = float(config["tf_fac"])

        # get flaring at r_planet
        self.hr_planet = self.hr * (self.r_planet / self.r_ref) ** (0.5 - self.q)

        # get length scale l
        self.l = (2/3) * self.hr_planet * self.r_planet

        # get sound speed at planet radius and reference radius
        v_kep_planet    = np.sqrt(self.G_const*self.m_star*self.m_solar / (self.r_planet*self.au))
        v_kep_ref       = np.sqrt(self.G_const*self.m_star*self.m_solar / (self.r_ref*self.au))
        self.c_s_planet = v_kep_planet*self.hr_planet
        self.c_s_0      = v_kep_ref*self.hr 

        # clockwise rotation factor
        #if self.cw_rotation == True:
        #    self.a_cw = -1
        #else:
        self.a_cw = 1

        # get height scale at reference radius
        self.h_ref = self.hr * self.r_ref

        # get flaring exponent beta (exponent for h, NOT h/r) for mcfost
        self.beta = 1 + 0.5 - self.q

        # get thermal mass
        self.m_thermal = 0.6666667 * self.hr_planet ** 3 * (self.m_star * self.m_solar / self.m_jupiter)

        # get dimensions
        if self.dimensionless:
            self.U_len  = self.r_ref * self.au
            self.U_mass = self.m_solar
            self.U_time = np.sqrt(self.U_len**3 / (self.G_const * self.U_mass)) 
            self.U_vel  = self.U_len / self.U_time / 1E5

    # checks that the combination of parameters the user has provided is sensical
    def _do_sanity_checks(self) -> bool:
        """Called to check that the parameters specified by the user will not break the results/run. Some parameter combinations are incompatible.
        """

        print("\n* Performing checks on model parameters:")

        if self.m_planet != None:
            print(f"M_thermal = {self.m_thermal:.3f} M_Jup")
            print(f"M_planet  = {(self.m_planet / self.m_thermal):.3f} M_th")

            # check that planet mass does not exceed thermal mass
            if self.m_planet > self.m_thermal:
                print("WARNING: M_planet > M_thermal. This may break the solution.")
                
        else:
            print(f"M_thermal = {self.m_thermal:.3f} M_Jup")
            print(f"M_planets = {np.array2string(np.array(self.m_planet_array)/self.m_thermal, precision=3, floatmode='fixed')} M_th")

            # check that planet mass does not exceed thermal mass
            exceed = False
            for mass in self.m_planet_array:
                if mass > self.m_thermal:
                    exceed = True
            if exceed ==  True:
                print("WARNING: At least one planet mass exceeds thermal mass. This may break the solution. ")

        # check grid type
        if self.grid_type != "cartesian" and self.grid_type != "cylindrical" and self.grid_type != "mcfost":
            raise Exception("Invalid grid type. Choose either cartesian or cylindrical or mcfost)")
        
        # check settings OK if mcfost is to be run
        if self.run_mcfost:
            if self.grid_type != "mcfost":
                raise Exception("Cannot run mcfost without using mcfost grid")
            elif self.write_FITS != True:
                raise Exception("Cannot run mcfost without writing FITS file (ie. require write_FITS: True)")

        # check linear box scale factor
        if self.scale_box != 1:
            print("WARNING: Changing linear box scale factor can cause strange results.")

        # check CFL
        if self.CFL > 0.5:
            print("WARNING: CFL chosen > 0.5, this will likely break the numerical PDE solver.")

        # check box warp
        if not self.box_warp:
            print("WARNING: Choosing box_warp=False may invalidate your results.")

        # check IC read out
        if self.use_box_IC:
            print("WARNING: Choosing use_box_IC=True will almost certainly invalidate your results.")

        # debug plots t eta
        if self.show_teta_debug_plots:
            print("WARNING: Choosing show_teta_debug_plots=True may cause the run to fail.")

        print("Parameters Ok - continuing")
        return True

# read in .yaml file, check keys correspond to parameters and return dictionary of parameters
def _load_config_file(config_file: str, default_config_dict: dict = None) -> dict:
    """Reads .yaml parameter file into a dictionary so that Wakeflow may parse it.
    """

    # read in config file as dictionary
    config_dict = yaml.load(open(config_file), Loader=yaml.FullLoader)

    # check that keys in config_dict correspond to actual Wakeflow parameters
    if default_config_dict is not None:
        for key in config_dict.keys():
            if key not in default_config_dict.keys():
                raise Exception(f"{key} is not a valid parameter.")

    return config_dict

# write dict to .yaml file
def _write_config_file(config_dict: dict, directory: str, filename: str) -> None:
    """Writes parameters dictionary to .yaml file.
    """

    with open(f'{directory}{filename}', 'w') as yaml_file:
        yaml.dump(config_dict, yaml_file, default_flow_style=False)

def _run_setup(param_dict: dict, overwrite: bool = False) -> _Parameters:
    """Perform setup by generating parameters object, checking the parameters are okay, creating results directory,
    writing .yaml file with parameters used to results dir, and calling MCFOST to generate grid data if needed.
    """

    params = _Parameters(param_dict)

    # do sanity checks and exit if not passed
    params._do_sanity_checks()

    # check if anything is being saved, if yes create directory
    if params.save_perturbations or params.save_total or params.make_midplane_plots:

        # check if directory for system exists, if not create it
        system_path = f"{params.system}/"
        os.makedirs(system_path, exist_ok=True)

        # create results directory path
        results_path = f"{params.system}/{params.name}/"

        # check if results directory already exists, ask to overwrite if yes
        results_exist = os.path.isdir(results_path)
        if results_exist is True:
            if overwrite:
                print("Overwriting previous results")
                sh.rmtree(results_path)
                os.makedirs(results_path)
            else:
                raise Exception("Results already exist for run name. Either choose a different name, or run with overwrite=True")

        # create individual directories for each planet mass result
        if params.m_planet_array is not None:
            for mass in params.m_planet_array:
                individual_result_path = f"{params.system}/{params.name}/{mass}Mj"
                os.makedirs(individual_result_path, exist_ok=True)
        else:
            individual_result_path = f"{params.system}/{params.name}/{params.m_planet}Mj"
            os.makedirs(individual_result_path, exist_ok=True)

        # write parameters used to a file in the results directory
        _write_config_file(param_dict, results_path, f"{params.name}_config.yaml")

    else:

        pass

    # run mcfost grid setup if needed
    if params.run_mcfost or params.grid_type == "mcfost":

        from .mcfost_interface import _make_mcfost_parameter_file, _make_mcfost_grid_data

        # make directory for mcfost outputs
        mcfost_path = f"{params.system}/{params.name}/mcfost/"
        os.makedirs(mcfost_path, exist_ok=True)

        # generate mcfost parameter file
        _make_mcfost_parameter_file(params)

        # generate mcfost grid data to run analytics on
        _make_mcfost_grid_data(params)

    return params


